import pandas as pd

import dash
from dash import html, dcc
import dash_bootstrap_components as dbc
from dash_extensions.javascript import Namespace
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from dash_tabulator import DashTabulator

from . import tools as T


ns = Namespace("myNamespace", "tabulator")

tabulator_options = {
    "groupBy": "Label",
    "selectable": True,
    "headerFilterLiveFilterDelay": 3000,
    "layout": "fitDataFill",
    "height": "900px",
}

downloadButtonType = {
    "css": "btn btn-primary",
    "text": "Export",
    "type": "csv",
    "filename": "Metadata",
}

clearFilterButtonType = {"css": "btn btn-outline-dark", "text": "Clear Filters"}

meta_table = html.Div(
    id="meta-table-container",
    style={"minHeight": 100, "margin": "0%"},
    children=[
        DashTabulator(
            id="meta-table",
            columns=T.gen_tabulator_columns(
                add_ms_file_col=True,
                add_color_col=True,
                add_peakopt_col=True,
                add_ms_file_active_col=True,
            ),
            options=tabulator_options,
            downloadButtonType=downloadButtonType,
            clearFilterButtonType=clearFilterButtonType,
        )
    ],
)

options = [
    {"label": "Batch", "value": "Batch"},
    {"label": "Label", "value": "Label"},
    {"label": "Color", "value": "Color"},
    {"label": "Type", "value": "Type"},
    {"label": "Concentration", "value": "Concentration"},
]

_label = "Metadata"

_layout = html.Div(
    [
        html.H3("Metadata"),
        dcc.Upload(
            id="meta-upload",
            children=html.Div(["Drag and Drop or ", html.A("Select Files")]),
            style={
                "width": "100%",
                "height": "60px",
                "lineHeight": "60px",
                "borderWidth": "1px",
                "borderStyle": "dashed",
                "borderRadius": "5px",
                "textAlign": "center",
                "margin": "10px",
            },
            # Allow multiple files to be uploaded
            multiple=True,
        ),
        dcc.Markdown("---"),
        dcc.Markdown("##### Actions"),
        dbc.Row(
            [
                dcc.Dropdown(
                    id="meta-action",
                    options=[
                        {"label": "Set", "value": "Set"},
                        {"label": "Create column", "value": "create_column"},
                        {"label": "Delete column", "value": "delete_column"},
                        # {'label': 'Delete selected files', 'value': 'delete_ms_files'},
                    ],
                    value="Set",
                    style={"width": "150px"},
                ),
                dcc.Dropdown(
                    id="meta-column",
                    options=options,
                    value=None,
                    style={"width": "150px"},
                ),
                dcc.Input(id="meta-input"),
                dcc.Dropdown(
                    id="meta-input-bool",
                    options=[
                        {"value": "True", "label": "True"},
                        {"value": "False", "label": "False"},
                    ],
                    value=None,
                ),
                html.Button("Apply", id="meta-apply"),
            ],
            style={"marginLeft": "5px"},
        ),
        dcc.Markdown("---"),
        dcc.Loading(meta_table),
        html.Div(id="meta-upload-output", style={"visibility": "hidden"}),
        html.Div(id="meta-apply-output", style={"visibility": "hidden"}),
    ]
)


_outputs = html.Div(
    id="meta-outputs",
    children=[
        html.Div(id={"index": "meta-apply-output", "type": "output"}),
        html.Div(id={"index": "meta-table-saved-on-edit-output", "type": "output"}),
    ],
)


def layout():
    return _layout


def callbacks(app, fsc, cache):
    @app.callback(
        Output("meta-table", "data"),
        Output("meta-table", "columns"),
        Output("meta-column", "options"),
        Input("tab", "value"),
        Input("meta-upload-output", "children"),
        Input("meta-apply-output", "children"),
        State("wdir", "children"),
    )
    def meta_load_table(tab, upload_trigger, apply_trigger, wdir):
        if tab != "Metadata":
            raise PreventUpdate
        df = T.get_metadata(wdir)
        columns = df.columns.to_list()
        options = [{"label": col, "value": col} for col in columns if col != "index"]
        return (
            df.to_dict("records"),
            T.gen_tabulator_columns(
                df.columns,
                add_ms_file_col=True,
                add_color_col=True,
                add_peakopt_col=True,
                add_ms_file_active_col=True,
            ),
            options,
        )

    @app.callback(
        Output("meta-upload-output", "children"),
        Input("meta-upload", "contents"),
        Input("meta-upload", "filename"),
        State("wdir", "children"),
    )
    def meta_upload(contents, filename, wdir):
        if contents is None:
            raise PreventUpdate
        df = T.get_metadata(wdir)
        contents = T.parse_table_content(contents[0], filename[0])
        df = T.merge_metadata(df, contents)
        if "index" not in df.columns:
            df = df.reset_index()
        T.write_metadata(df, wdir)
        return "Table updated"

    @app.callback(
        Output("meta-table", "downloadButtonType"),
        Input("tab", "value"),
        State("active-workspace", "children"),
    )
    def update_table_export_fn(tab, ws_name):
        fn = f"{T.today()}__MINT__{ws_name}__metadata"
        downloadButtonType = {
            "css": "btn btn-primary",
            "text": "Export",
            "type": "csv",
            "filename": fn,
        }
        return downloadButtonType

    @app.callback(
        Output({"index": "meta-apply-output", "type": "output"}, "children"),
        Output("meta-apply-output", "children"),
        Input("meta-apply", "n_clicks"),
        State("meta-table", "data"),
        State("meta-table", "multiRowsClicked"),
        State("meta-table", "dataFiltered"),
        State("meta-action", "value"),
        State("meta-column", "value"),
        State("meta-input", "value"),
        State("meta-input-bool", "value"),
        State("wdir", "children"),
    )
    def meta_apply(
        n_clicks,
        data,
        selected_rows,
        data_filtered,
        action,
        column,
        value,
        value_bool,
        wdir,
    ):
        """
        This callback saves applies the column actions and
        saves the table to the harddrive.
        """
        if (n_clicks is None) or (data is None) or (len(data) == 0):
            raise PreventUpdate
        if action == "Set" and column == "PeakOpt":
            value = value_bool
        df = pd.DataFrame(data)
        if "index" in df.columns:
            df = df.set_index("index")
        else:
            df = df.reset_index()
        prop_id = dash.callback_context.triggered[0]["prop_id"]
        if prop_id == "meta-apply.n_clicks":
            if action == "Set":
                filtered_rows = [r for r in data_filtered["rows"] if r is not None]
                filtered_ndx = [r["index"] for r in filtered_rows]
                if selected_rows == []:
                    # If nothing is selected apply to all visible rows
                    ndxs = filtered_ndx
                else:
                    # If something is selected only apply to selected rows
                    ndxs = [
                        r["index"] for r in selected_rows if r["index"] in filtered_ndx
                    ]
                if len(ndxs) == 0 or column is None:
                    # If no column is selected just save the table.
                    T.write_metadata(df, wdir)
                    return dbc.Alert("Metadata saved.", color="info"), "Applied"
                df.loc[ndxs, column] = value
            elif action == "create_column":
                df[value] = ""
            elif action == "delete_column":
                del df[column]
        T.write_metadata(df, wdir)
        return dbc.Alert("Metadata saved.", color="info"), "Applied"

    @app.callback(
        Output(
            {"index": "meta-table-saved-on-edit-output", "type": "output"}, "children"
        ),
        Input("meta-table", "cellEdited"),
        State("meta-table", "data"),
        State("wdir", "children"),
    )
    def save_table_on_edit(cell_edited, data, wdir):
        """
        This callback saves the table on cell edits.
        This saves some bandwidth.
        """
        if data is None or cell_edited is None:
            raise PreventUpdate
        df = pd.DataFrame(data)
        T.write_metadata(df, wdir)
        return dbc.Alert("Metadata saved.", color="info")

    @app.callback(
        Output("meta-input", "style"),
        Output("meta-input-bool", "style"),
        Input("meta-action", "value"),
        Input("meta-column", "value"),
    )
    def set_peak_opt_show_boolean_dropdown(action, column):
        visible = {"visibility": "visible", "width": "150px", "margin": 0}
        hidden = {"visibility": "hidden", "width": "0px", "margin": 0}
        if (action == "Set") & (column == "PeakOpt"):
            return hidden, visible
        return visible, hidden



from tqdm import tqdm

import matplotlib

matplotlib.use("Agg")
from matplotlib import pyplot as plt
import seaborn as sns

from dash import html, dcc
import dash_bootstrap_components as dbc
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State

from . import tools as T

graph_options = [
    {"label": "Histograms", "value": "hist"},
    {"label": "Boxplots", "value": "boxplot"},
    {"label": "Probability Density", "value": "density"},
]

_label = "Distributions"

_layout = html.Div(
    [
        html.H3("Quality Control"),
        html.Button("Update", id="dist-update"),
        dcc.Dropdown(
            id="dist-graphs",
            options=graph_options,
            value=["hist"],
            multi=True,
            placeholder="Kinds of graphs",
        ),
        dcc.Checklist(
            id="dist-select",
            options=[{"label": "Dense", "value": "Dense"}],
            value=["Dense"],
        ),
        html.Div(id="dist-figures", style={"float": "center"}),
    ]
)


layout_no_data = html.Div(
    [
        dcc.Markdown(
            """### No results generated yet. 
    MINT has not been run yet. The Quality Control tabs uses the processed data. 
    To generate it please add MS-files as well as a valid peaklist. 
    Then execute MINT data processing routine witby clicking the `RUN MINT` button. 
    Once results have been produced you can access the QC tools."""
        ),
    ]
)


def layout():
    return _layout


def callbacks(app, fsc, cache):
    @app.callback(
        Output("dist-figures", "children"),
        Input("dist-update", "n_clicks"),
        State("tab", "value"),
        State("ana-groupby", "value"),
        State("dist-graphs", "value"),
        State("dist-select", "value"),
        State("ana-file-types", "value"),
        State("ana-peak-labels-include", "value"),
        State("ana-peak-labels-exclude", "value"),
        State("wdir", "children"),
    )
    def qc_figures(
        n_clicks,
        tab,
        groupby,
        kinds,
        options,
        file_types,
        include_labels,
        exclude_labels,
        wdir,
    ):

        if n_clicks is None:
            raise PreventUpdate

        df = T.get_complete_results(
            wdir,
            include_labels=include_labels,
            exclude_labels=exclude_labels,
            file_types=file_types,
        )

        if len(df) == 0:
            return "No results yet. First run MINT."

        if "boxplot" in kinds:
            if groupby is None or len(groupby) == 0:
                return dbc.Alert(
                    'For boxplots a "Group-by" column has to be set', color="info"
                )
            if len(df[groupby].drop_duplicates()) <= 1:
                return dbc.Alert(
                    'For boxplots at least two groups have to be defined in selected "Group-by" column in metadata sheet.',
                    color="info",
                )

        sns.set_context("paper")

        sort_by_col = "Batch"
        quant_col = "log(peak_max+1)"

        if sort_by_col is not None:
            df = df.sort_values(["peak_label", sort_by_col])

        if options is None:
            options = []

        figures = []
        n_total = len(df.peak_label.drop_duplicates())
        for i, (peak_label, grp) in tqdm(
            enumerate(df.groupby("peak_label")), total=n_total
        ):

            if not "Dense" in options:
                figures.append(
                    dcc.Markdown(f"#### `{peak_label}`", style={"float": "center"})
                )
            fsc.set("progress", int(100 * (i + 1) / n_total))

            # Sorting to ensure similar legends
            if sort_by_col is not None:
                grp = grp.sort_values(sort_by_col).reset_index(drop=True)

            # if len(grp) < 1:
            #    continue

            if "hist" in kinds:
                sns.displot(data=grp, x=quant_col, height=3, hue=groupby, aspect=1)
                plt.title(peak_label)
                fig_label = f"by-{groupby}__{quant_col}__{peak_label}"
                T.savefig(kind="hist", wdir=wdir, label=fig_label)
                src = T.fig_to_src(dpi=150)
                figures.append(html.Img(src=src, style={"width": "300px"}))

            if "density" in kinds:
                sns.displot(
                    data=grp,
                    x=quant_col,
                    hue=groupby,
                    kind="kde",
                    common_norm=False,
                    height=3,
                    aspect=1,
                )
                plt.title(peak_label)
                fig_label = f"by-{groupby}__{quant_col}__{peak_label}"
                T.savefig(kind="density", wdir=wdir, label=fig_label)
                src = T.fig_to_src(dpi=150)
                figures.append(html.Img(src=src, style={"width": "300px"}))

            if "boxplot" in kinds:
                n_groups = len(grp[groupby].drop_duplicates())
                aspect = max(1, n_groups / 10)
                sns.catplot(
                    data=grp,
                    y=quant_col,
                    x=groupby,
                    height=3,
                    kind="box",
                    aspect=aspect,
                    color="w",
                )
                if quant_col in ["peak_max", "peak_area"]:
                    plt.gca().ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
                plt.title(peak_label)
                plt.xticks(rotation=90)
                fig_label = f"by-{groupby}__{quant_col}__{peak_label}"
                T.savefig(kind="boxplot", wdir=wdir, label=fig_label)
                src = T.fig_to_src(dpi=150)
                figures.append(html.Img(src=src, style={"width": "300px"}))

            if not "Dense" in options:
                figures.append(dcc.Markdown("---"))

            # if i == 3: break

        return figures



from dash import html, dcc
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from ms_mint.notebook import Mint

from . import tools as T


_label = "Hierachical Clustering"

_options = ["Transposed"]

options = [{"value": x, "label": x} for x in _options]

_layout = html.Div(
    [
        html.H3(_label),
        dcc.Input(
            id="hc-figsize-x", placeholder="Figure size x", value=8, type="number"
        ),
        dcc.Input(
            id="hc-figsize-y", placeholder="Figure size x", value=8, type="number"
        ),
        dcc.Dropdown(id="hc-options", options=options, value=[]),
        html.Button("Update", id="hc-update"),
        dcc.Loading(
            html.Div(
                id="hc-figures",
                style={
                    "margin": "auto",
                    "text-align": "center",
                    "maxWidth": "100%",
                    "minHeight": "300px",
                },
            )
        ),
    ]
)

_ouptuts = html.Div([])


def layout():
    return _layout


def callbacks(app, fsc, cache):
    @app.callback(
        Output("hc-figures", "children"),
        Input("hc-update", "n_clicks"),
        State("hc-figsize-x", "value"),
        State("hc-figsize-y", "value"),
        State("hc-options", "value"),
        State("ana-file-types", "value"),
        State("ana-peak-labels-include", "value"),
        State("ana-peak-labels-exclude", "value"),
        State("ana-ms-order", "value"),
        State("wdir", "children"),
    )
    def create_figure(
        n_clicks,
        fig_size_x,
        fig_size_y,
        options,
        file_types,
        include_labels,
        exclude_labels,
        ms_order,
        wdir,
    ):

        if n_clicks is None:
            raise PreventUpdate

        if options is None: 
            options = []
        
        print(options)

        mint = Mint()

        if fig_size_x is None:
            fig_size_x = 8
        if fig_size_y is None:
            fig_size_y = 8

        fig_size_x = min(float(fig_size_x), 100)
        fig_size_y = min(float(fig_size_y), 100)

        df = T.get_complete_results(
            wdir,
            include_labels=include_labels,
            exclude_labels=exclude_labels,
            file_types=file_types,
        )

        df["ms_file"] = df["MS-file"]
        mint.results = df

        mint.plot.hierarchical_clustering(
            figsize=(fig_size_x, fig_size_y), transposed="Transposed" in options
        )

        src = T.fig_to_src()

        return html.Img(src=src, style={"maxWidth": "80%"})



import numpy as np
import seaborn as sns
import plotly.express as px

from dash import html, dcc
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

import dash_bootstrap_components as dbc

from ms_mint import Mint

from . import tools as T

options = [{"value": i, "label": i} for i in ["Standard", "Corner"]]

_layout = html.Div(
    [
        html.H3("Principal Components Analysis"),
        html.Button("Run PCA", id="pca-update"),
        dcc.Dropdown(
            id="pca-options",
            options=options,
            value=["Standard"],
            multi=True,
            placeholder="Scaling used before PCA",
        ),
        html.Label("Number of PCA components"),
        dcc.Slider(
            id="pca-nvars",
            value=3,
            min=2,
            max=10,
            step=1,
            marks={i: f"{i}" for i in range(2, 11)},
        ),
        html.Label("Height of facets"),
        dcc.Slider(
            id="pca-facent-height",
            value=2.5,
            min=1,
            max=5,
            step=0.1,
            marks={i: f"{i}" for i in np.arange(1, 5.5, 0.5)},
        ),
        html.H4("Cumulative explained variance"),
        dcc.Loading(
            html.Div(
                id="pca-figure-explained-variance",
                style={"margin": "auto", "text-align": "center"},
            )
        ),
        html.H4("Scatter plot of principal components"),
        dcc.Loading(
            html.Div(
                id="pca-figure-pairplot",
                style={"margin": "auto", "text-align": "center"},
            )
        ),
        html.H4("Principal components compositions"),
        dcc.Loading(dcc.Graph(id="pca-figure-contrib")),
    ]
)

_label = "PCA"


def layout():
    return _layout


def callbacks(app, fsc, cache):
    @app.callback(
        Output("pca-figure-explained-variance", "children"),
        Output("pca-figure-pairplot", "children"),
        Output("pca-figure-contrib", "figure"),
        Input("pca-update", "n_clicks"),
        State("pca-nvars", "value"),
        State("pca-facent-height", "value"),
        State("ana-groupby", "value"),
        State("ana-peak-labels-include", "value"),
        State("ana-peak-labels-exclude", "value"),
        State("ana-normalization-cols", "value"),
        State("ana-file-types", "value"),
        State("pca-options", "value"),
        State("wdir", "children"),
    )
    def create_pca(
        n_clicks,
        n_components,
        facet_height,
        groupby,
        include_labels,
        exclude_labels,
        norm_cols,
        file_types,
        options,
        wdir,
    ):
        if n_clicks is None:
            raise PreventUpdate
        if norm_cols is None:
            norm_cols = []

        df = T.get_complete_results(
            wdir,
            include_labels=include_labels,
            exclude_labels=exclude_labels,
            file_types=file_types,
        )

        if file_types is not None and len(file_types) > 0:
            df = df[df["Type"].isin(file_types)]

        if groupby is not None and len(groupby) > 0:
            labels = (
                df[["ms_file", groupby]].drop_duplicates().set_index("ms_file")
            )
        else:
            labels = None
            groupby = None

        if len(norm_cols) != 0:
            if ("peak_label" in norm_cols) and ("ms_file" in norm_cols):
                return dbc.Alert(
                    "'peak_label' and 'ms_file' should not be used together for normalization!",
                    color="danger",
                )

            df = df[df.Batch.notna()]
            cols = ["peak_max"]
            df.loc[:, cols] = (
                (
                    df[cols]
                    - df[cols + norm_cols]
                    .groupby(norm_cols)
                    .transform("median")[cols]
                    .values
                )
                / df[cols + norm_cols].groupby(norm_cols).transform("std")[cols].values
            ).reset_index()

        figures = []
        mint = Mint()
        mint.results = df
        mint.pca.run(n_components=n_components)

        ndx = mint.pca.results["df_projected"].index.to_list()
        mint.pca.plot.cumulative_variance()

        src = T.fig_to_src()

        figures.append(html.Img(src=src))

        if labels is not None:
            labels = labels.loc[ndx].values

        with sns.plotting_context("paper"):
            mint.pca.plot.pairplot(
                labels=labels,
                n_vars=n_components,
                height=facet_height,
                corner="Corner" in options,
            )

        src = T.fig_to_src()
        figures.append(html.Img(src=src))

        contrib = mint.pca.results["feature_contributions"]

        fig_contrib = px.bar(
            data_frame=contrib,
            x="peak_label",
            y="Coefficient",
            facet_col="PC",
            facet_col_wrap=1,
            height=200 * n_components + 200,
        )

        fig_contrib.update_layout(autosize=True)

        return figures[0], figures[1], fig_contrib



import os
import io
import shutil
import base64
import subprocess
import platform
import logging

import numpy as np
import pandas as pd

from tqdm import tqdm
from glob import glob
from pathlib import Path as P

import urllib3, ftplib
from urllib.parse import urlparse
from bs4 import BeautifulSoup

import matplotlib as mpl

mpl.use("Agg")

from matplotlib import pyplot as plt
import matplotlib.cm as cm

import ms_mint
from ms_mint.io import ms_file_to_df
from ms_mint.targets import standardize_targets, read_targets
from ms_mint.io import convert_ms_file_to_feather
from ms_mint.standards import TARGETS_COLUMNS

from datetime import date

from .filelock import FileLock


def lock(fn):
    return FileLock(f"{fn}.lock", timeout=1)


def today():
    return date.today().strftime("%y%m%d")


def get_versions():
    string = ""
    try:
        string += subprocess.getoutput("conda env export --no-build")
    except:
        pass
    return string


def get_issue_text():
    return f"""
    %0A%0A%0A%0A%0A%0A%0A%0A%0A
    MINT version: {ms_mint.__version__}%0A
    OS: {platform.platform()}%0A
    Versions: ---
    """


def parse_ms_files(contents, filename, date, target_dir):
    content_type, content_string = contents.split(",")
    decoded = base64.b64decode(content_string)
    fn_abs = os.path.join(target_dir, filename)
    with lock(fn_abs):
        with open(fn_abs, "wb") as file:
            file.write(decoded)
    new_fn = convert_ms_file_to_feather(fn_abs)
    if os.path.isfile(new_fn):
        os.remove(fn_abs)


def parse_pkl_files(contents, filename, date, target_dir, ms_mode=None):
    content_type, content_string = contents.split(",")
    decoded = base64.b64decode(content_string)
    df = pd.read_csv(io.StringIO(decoded.decode("utf-8")))
    df = standardize_targets(df, ms_mode=ms_mode)
    df = df.drop_duplicates()
    return df


def get_dirnames(path):
    dirnames = [f.name for f in os.scandir(path) if f.is_dir()]
    return dirnames


def workspace_path(tmpdir, ws_name):
    return os.path.join(tmpdir, "workspaces", ws_name)


def maybe_migrate_workspaces(tmpdir):
    if not P(tmpdir).is_dir():
        return None
    dir_names = get_dirnames(tmpdir)
    ws_path = get_workspaces_path(tmpdir)
    if not os.path.isdir(ws_path) and len(dir_names) > 0:
        logging.info("Migrating to new directory scheme.")
        os.makedirs(ws_path)
        for dir_name in dir_names:
            old_dir = os.path.join(tmpdir, dir_name)
            new_dir = workspace_path(tmpdir, dir_name)
            shutil.move(old_dir, new_dir)
            logging.info("Moving", old_dir, "to", new_dir)


def maybe_update_workpace_scheme(wdir):
    old_pkl_fn = P(wdir) / "peaklist" / "peaklist.csv"
    new_pkl_fn = P(get_targets_fn(wdir))
    new_path = new_pkl_fn.parent
    old_path = old_pkl_fn.parent
    if old_pkl_fn.is_file():
        logging.info(f"Moving targets file to new default location ({new_pkl_fn}).")
        if not new_path.is_dir():
            os.makedirs(new_path)
        os.rename(old_pkl_fn, new_pkl_fn)
        shutil.rmtree(old_path)


def workspace_exists(tmpdir, ws_name):
    path = workspace_path(tmpdir, ws_name)
    return os.path.isdir(path)


def get_active_workspace(tmpdir):
    """Returns name of last activated workspace,
    if workspace still exists. Otherwise,
    return None.
    """
    fn_ws_info = os.path.join(tmpdir, ".active-workspace")
    if not os.path.isfile(fn_ws_info):
        return None
    with open(fn_ws_info, "r") as file:
        ws_name = file.read()
    if ws_name in get_workspaces(tmpdir):
        return ws_name
    else:
        return None


def save_activated_workspace(tmpdir, ws_name):
    fn_ws_info = os.path.join(tmpdir, ".active-workspace")
    with open(fn_ws_info, "w") as file:
        file.write(ws_name)


def create_workspace(tmpdir, ws_name):
    path = workspace_path(tmpdir, ws_name)
    assert not os.path.isdir(path)
    os.makedirs(path)
    os.makedirs(os.path.join(path, "ms_files"))
    os.makedirs(os.path.join(path, "targets"))
    os.makedirs(os.path.join(path, "results"))
    os.makedirs(os.path.join(path, "figures"))
    os.makedirs(os.path.join(path, "chromato"))


def get_workspaces_path(tmpdir):
    # Defines the path to the workspaces
    # relative to `tmpdir`
    return os.path.join(tmpdir, "workspaces")


def get_workspaces(tmpdir):
    ws_path = get_workspaces_path(tmpdir)
    if not P(ws_path).is_dir():
        return []
    ws_names = get_dirnames(ws_path)
    ws_names = [ws for ws in ws_names if not ws.startswith(".")]
    ws_names.sort()
    return ws_names


class Chromatograms:
    def __init__(self, wdir, targets, ms_files, progress_callback=None):
        self.wdir = wdir
        self.targets = targets
        self.ms_files = ms_files
        self.n_peaks = len(targets)
        self.n_files = len(ms_files)
        self.progress_callback = progress_callback

    def create_all(self):
        for fn in tqdm(self.ms_files):
            self.create_all_for_ms_file(fn)
        return self

    def create_all_for_ms_file(self, ms_file: str):
        fn = ms_file
        df = ms_file_to_df(fn)
        for ndx, row in self.targets.iterrows():
            mz_mean, mz_width = row[["mz_mean", "mz_width"]]
            fn_chro = get_chromatogram_fn(fn, mz_mean, mz_width, self.wdir)
            if os.path.isfile(fn_chro):
                continue
            dirname = os.path.dirname(fn_chro)
            if not os.path.isdir(dirname):
                os.makedirs(dirname)
            dmz = mz_mean * 1e-6 * mz_width
            chrom = df[(df["mz"] - mz_mean).abs() <= dmz]
            chrom["scan_time"] = chrom["scan_time"].round(3)
            chrom = chrom.groupby("scan_time").max().reset_index()
            chrom[["scan_time", "intensity"]].to_feather(fn_chro)

    def get_single(self, mz_mean, mz_width, ms_file):
        return get_chromatogram(ms_file, mz_mean, mz_width, self.wdir)


def create_chromatograms(ms_files, targets, wdir):
    for fn in tqdm(ms_files):
        fn_out = os.path.basename(fn)
        fn_out, _ = os.path.splitext(fn_out)
        fn_out += ".feather"
        for ndx, row in targets.iterrows():
            mz_mean, mz_width = row[["mz_mean", "mz_width"]]
            fn_chro = get_chromatogram_fn(fn, mz_mean, mz_width, wdir)
            if not os.path.isfile(fn_chro):
                create_chromatogram(fn, mz_mean, mz_width, fn_chro)


def create_chromatogram(ms_file, mz_mean, mz_width, fn_out, verbose=False):
    if verbose:
        print("Creating chromatogram")
    df = ms_file_to_df(ms_file)
    if verbose:
        print("...file read")
    dirname = os.path.dirname(fn_out)
    if not os.path.isdir(dirname):
        os.makedirs(dirname)
    dmz = mz_mean * 1e-6 * mz_width
    chrom = df[(df["mz"] - mz_mean).abs() <= dmz]
    chrom["scan_time"] = chrom["scan_time"].round(3)
    chrom = chrom.groupby("scan_time").max().reset_index()
    with lock(fn_out):
        chrom[["scan_time", "intensity"]].to_feather(fn_out)
    if verbose:
        print("...done creating chromatogram.")
    return chrom


def get_chromatogram(ms_file, mz_mean, mz_width, wdir):
    fn = get_chromatogram_fn(ms_file, mz_mean, mz_width, wdir)
    if not os.path.isfile(fn):
        chrom = create_chromatogram(ms_file, mz_mean, mz_width, fn)
    else:
        try:
            chrom = pd.read_feather(fn)
        except:
            os.remove(fn)
            logging.warning(f"Cound not read {fn}.")
            return None

    chrom = chrom.rename(
        columns={
            "retentionTime": "scan_time",
            "intensity array": "intensity",
            "m/z array": "mz",
        }
    )

    return chrom


def get_chromatogram_fn(ms_file, mz_mean, mz_width, wdir):
    ms_file = os.path.basename(ms_file)
    base, _ = os.path.splitext(ms_file)
    fn = (
        os.path.join(wdir, "chromato", f"{mz_mean}-{mz_width}".replace(".", "_"), base)
        + ".feather"
    )
    return fn


def get_targets_fn(wdir):
    return os.path.join(wdir, "targets", "targets.csv")


def get_targets(wdir):
    fn = get_targets_fn(wdir)
    if os.path.isfile(fn):
        targets = read_targets(fn).set_index("peak_label")
    else:
        targets = pd.DataFrame(columns=TARGETS_COLUMNS)
    return targets


def update_targets(wdir, peak_label, rt_min=None, rt_max=None, rt=None):
    targets = get_targets(wdir)

    if isinstance(peak_label, str):
        if rt_min is not None and not np.isnan(rt_min):
            targets.loc[peak_label, "rt_min"] = rt_min
        if rt_max is not None and not np.isnan(rt_max):
            targets.loc[peak_label, "rt_max"] = rt_max
        if rt is not None and not np.isnan(rt):
            targets.loc[peak_label, "rt"] = rt

    if isinstance(peak_label, int):
        targets = targets.reset_index()
        if rt_min is not None and not np.isnan(rt_min):
            targets.loc[peak_label, "rt_min"] = rt_min
        if rt_max is not None and not np.isnan(rt_max):
            targets.loc[peak_label, "rt_max"] = rt_max
        if rt is not None and not np.isnan(rt):
            targets.loc[peak_label, "rt"] = rt
        targets = targets.set_index("peak_label")

    fn = get_targets_fn(wdir)
    with lock(fn):
        targets.to_csv(fn)


def get_results_fn(wdir):
    return os.path.join(wdir, "results", "results.csv")


def get_results(wdir):
    fn = get_results_fn(wdir)
    df = pd.read_csv(fn)
    df["MS-file"] = [filename_to_label(fn) for fn in df["ms_file"]]
    return df


def get_metadata(wdir):
    fn = get_metadata_fn(wdir)
    fn_path = os.path.dirname(fn)
    ms_files = get_ms_fns(wdir, abs_path=False)
    ms_files = [filename_to_label(fn) for fn in ms_files]
    df = None
    if not os.path.isdir(fn_path):
        os.makedirs(fn_path)
    if os.path.isfile(fn):
        df = pd.read_csv(fn)
        if "MS-file" not in df.columns:
            df = None

    if df is None or len(df) == 0:
        df = init_metadata(ms_files)

    for col in [
        "Color",
        "Column",
        "Row",
        "Batch",
        "Label",
        "InAnalysis",
        "PeakOpt",
        "MS-file",
    ]:
        if col not in df.columns:
            df[col] = None

    df = df[df["MS-file"] != ""]

    df = df.groupby("MS-file").first().reindex(ms_files).reset_index()

    if "PeakOpt" not in df.columns:
        df["PeakOpt"] = False

    else:
        df["PeakOpt"] = df["PeakOpt"].astype(bool)

    if "InAnalysis" not in df.columns:
        df["InAnalysis"] = True
    else:
        df["InAnalysis"] = df["InAnalysis"].astype(bool)

    if "index" in df.columns:
        del df["index"]

    df["Column"] = df["Column"].apply(format_columns)

    df["Type"] = df["Type"].fillna("Not set")

    df.reset_index(inplace=True)

    return df


def init_metadata(ms_files):
    ms_files = list(ms_files)
    ms_files = [filename_to_label(fn) for fn in ms_files]
    df = pd.DataFrame({"MS-file": ms_files})
    df["InAnalysis"] = True
    df["Label"] = ""
    df["Color"] = None
    df["Type"] = "Biological Sample"
    df["RunOrder"] = ""
    df["Batch"] = ""
    df["Row"] = ""
    df["Column"] = ""
    df["PeakOpt"] = ""
    return df


def write_metadata(meta, wdir):
    fn = get_metadata_fn(wdir)
    with lock(fn):
        meta.to_csv(fn, index=False)


def get_metadata_fn(wdir):
    fn = os.path.join(wdir, "metadata", "metadata.csv")
    return fn


def get_ms_dirname(wdir):
    return os.path.join(wdir, "ms_files")


def get_ms_fns(wdir, abs_path=True):
    path = get_ms_dirname(wdir)
    fns = glob(os.path.join(path, "**", "*.*"), recursive=True)
    fns = [fn for fn in fns if is_ms_file(fn)]
    if not abs_path:
        fns = [os.path.basename(fn) for fn in fns]
    return fns


def is_ms_file(fn: str):
    if (
        fn.lower().endswith(".mzxml")
        or fn.lower().endswith(".mzml")
        or fn.lower().endswith(".feather")
    ):
        return True
    return False


def Basename(fn):
    fn = os.path.basename(fn)
    fn, _ = os.path.splitext(fn)
    return fn


def format_columns(x):
    try:
        if (x is None) or (x == "") or np.isnan(x):
            return None
    except:
        print(type(x))
        print(x)
        assert False
    return f"{int(x):02.0f}"


def get_complete_results(
    wdir,
    include_labels=None,
    exclude_labels=None,
    file_types=None,
    include_excluded=False,
):
    meta = get_metadata(wdir)
    resu = get_results(wdir)

    if not include_excluded:
        meta = meta[meta["InAnalysis"]]
    df = pd.merge(meta, resu, on=["MS-file"])
    if include_labels is not None and len(include_labels) > 0:
        df = df[df.peak_label.isin(include_labels)]
    if exclude_labels is not None and len(exclude_labels) > 0:
        df = df[~df.peak_label.isin(exclude_labels)]
    if file_types is not None and file_types != []:
        df = df[df.Type.isin(file_types)]
    df["log(peak_max+1)"] = df.peak_max.apply(np.log1p)
    if "index" in df.columns:
        df = df.drop("index", axis=1)
    return df


def gen_tabulator_columns(
    col_names=None,
    add_ms_file_col=False,
    add_color_col=False,
    add_peakopt_col=False,
    add_ms_file_active_col=False,
    col_width="12px",
    editor="input",
):

    if col_names is None:
        col_names = []
    col_names = list(col_names)

    standard_columns = [
        "MS-file",
        "InAnalysis",
        "Color",
        "index",
        "PeakOpt",
    ]

    for col in standard_columns:
        if col in col_names:
            col_names.remove(col)

    columns = [
        {
            "formatter": "rowSelection",
            "titleFormatter": "rowSelection",
            "titleFormatterParams": {
                "rowRange": "active"  # only toggle the values of the active filtered rows
            },
            "hozAlign": "center",
            "headerSort": False,
            "width": "1px",
            "frozen": True,
        }
    ]

    if add_ms_file_col:
        columns.append(
            {
                "title": "MS-file",
                "field": "MS-file",
                "headerFilter": True,
                "headerSort": True,
                "editor": "input",
                "sorter": "string",
                "frozen": True,
            }
        )

    if add_color_col:
        columns.append(
            {
                "title": "Color",
                "field": "Color",
                "headerFilter": False,
                "editor": "input",
                "formatter": "color",
                "width": "3px",
                "headerSort": False,
            }
        )

    if add_peakopt_col:
        columns.append(
            {
                "title": "PeakOpt",
                "field": "PeakOpt",
                "headerFilter": False,
                "formatter": "tickCross",
                "width": "6px",
                "headerSort": True,
                "hozAlign": "center",
                "editor": True,
            }
        )

    if add_ms_file_active_col:
        columns.append(
            {
                "title": "InAnalysis",
                "field": "InAnalysis",
                "headerFilter": True,
                "formatter": "tickCross",
                "width": "6px",
                "headerSort": True,
                "hozAlign": "center",
                "editor": True,
            }
        )

    for col in col_names:
        content = {
            "title": col,
            "field": col,
            "headerFilter": True,
            "width": col_width,
            "editor": editor,
        }

        columns.append(content)
    return columns


def parse_table_content(content, filename):
    content_type, content_string = content.split(",")
    decoded = base64.b64decode(content_string)
    if filename.lower().endswith(".csv"):
        df = pd.read_csv(io.StringIO(decoded.decode("utf-8")))
    elif filename.lower().endswith(".xlsx"):
        df = pd.read_excel(io.BytesIO(decoded))
    return df


def fig_to_src(dpi=100):
    out_img = io.BytesIO()

    plt.savefig(out_img, format="jpeg", bbox_inches="tight", dpi=dpi)

    plt.close("all")
    out_img.seek(0)  # rewind file
    encoded = base64.b64encode(out_img.read()).decode("ascii").replace("\n", "")
    return "data:image/png;base64,{}".format(encoded)


def merge_metadata(old, new):
    old = old.set_index("MS-file")

    new = new.groupby("MS-file").first().replace("null", None)

    for col in new.columns:
        if col == "" or col.startswith("Unnamed"):
            continue
        if not col in old.columns:
            old[col] = None
        for ndx in new.index:
            value = new.loc[ndx, col]
            if value is None:
                continue
            if ndx in old.index:
                old.loc[ndx, col] = value
    return old.reset_index()


def file_colors(wdir):
    meta = get_metadata(wdir)
    colors = {}
    for ndx, (fn, co) in meta[["MS-file", "Color"]].iterrows():
        if not (isinstance(co, str)):
            co = None
        colors[fn] = co
    return colors


def get_figure_fn(kind, wdir, label, format):
    path = os.path.join(wdir, "figures", kind)
    clean_label = clean_string(label)
    fn = f"{kind}__{clean_label}.{format}"
    fn = os.path.join(path, fn)
    return path, fn


def clean_string(fn: str):
    for x in ['"', "'", "(", ")", "[", "]", " ", "\\", "/", "{", "}"]:
        fn = fn.replace(x, "_")
    return fn


def savefig(kind=None, wdir=None, label=None, format="png", dpi=150):
    path, fn = get_figure_fn(kind=kind, wdir=wdir, label=label, format=format)
    maybe_create(path)
    try:
        with lock(fn):
            plt.savefig(fn, dpi=dpi, bbox_inches="tight")
    except:
        print(f"Could not save figure {fn}, maybe no figure was created: {label}")
    return fn


def maybe_create(dir_name):
    if not os.path.isdir(dir_name):
        os.makedirs(dir_name)


def png_fn_to_src(fn):
    encoded_image = base64.b64encode(open(fn, "rb").read())
    return "data:image/png;base64,{}".format(encoded_image.decode())


def get_ms_fns_for_peakopt(wdir):
    """Extract the filenames for peak optimization from
    the metadata table and recreate the complete filename."""
    df = get_metadata(wdir)
    fns = df[df.PeakOpt.astype(bool) == True]["MS-file"]
    ms_files = get_ms_fns(wdir)
    mapping = {filename_to_label(fn): fn for fn in ms_files}
    fns = [mapping[fn] for fn in fns]
    return fns


def float_to_color(x, vmin=0, vmax=2, cmap=None):
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
    m = cm.ScalarMappable(norm=norm, cmap=cmap)
    return m.to_rgba(x)


def write_targets(targets, wdir):
    fn = get_targets_fn(wdir)
    if "peak_label" in targets.columns:
        targets = targets.set_index("peak_label")
    with lock(fn):
        targets.to_csv(fn)


def filename_to_label(fn: str):
    if is_ms_file(fn):
        fn = os.path.splitext(fn)[0]
    return os.path.basename(fn)


def get_filenames_from_url(url):
    if url.startswith("ftp"):
        return get_filenames_from_ftp_directory(url)
    if "://" in url:
        url = url.split("://")[1]
    with urllib3.PoolManager() as http:
        r = http.request("GET", url)
    soup = BeautifulSoup(r.data, "html")
    files = [A["href"] for A in soup.find_all("a", href=True)]
    return files


def get_filenames_from_ftp_directory(url):
    url_parts = urlparse(url)
    domain = url_parts.netloc
    path = url_parts.path
    ftp = ftplib.FTP(domain)
    ftp.login()
    ftp.cwd(path)
    filenames = ftp.nlst()
    ftp.quit()
    return filenames


def import_from_local_path(path, target_dir, fsc=None):
    fns = glob(os.path.join(path, "**", "*.*"), recursive=True)
    fns = [fn for fn in fns if is_ms_file(fn)]
    fns_out = []
    n_files = len(fns)
    for i, fn in enumerate(tqdm(fns)):
        if fsc is not None:
            fsc.set("progress", int(100 * (1 + i) / n_files))
        fn_out = P(target_dir) / P(fn).with_suffix(".feather").name
        if P(fn_out).is_file():
            continue
        fns_out.append(fn_out)
        try:
            convert_ms_file_to_feather(fn, fn_out)
        except:
            logging.warning(f"Could not convert {fn}")
    return fns_out


def df_to_in_memory_csv_file(df):
    buffer = io.StringIO()
    df.to_csv(buffer)
    buffer.seek(0)
    return buffer.getvalue


def df_to_in_memory_excel_file(df):
    def to_xlsx(bytes_io):
        xslx_writer = pd.ExcelWriter(bytes_io, engine="xlsxwriter")
        df.to_excel(xslx_writer, index=True, sheet_name="sheet1")
        xslx_writer.save()

    return to_xlsx


def has_na(df):
    return df.isna().sum().sum() > 0



#!/usr/bin/env python

import os
import sys
import subprocess
import multiprocessing
import argparse
import pkg_resources
import xlsxwriter

from waitress import serve
from os.path import expanduser
from pathlib import Path as P
from collections import namedtuple
from multiprocessing import freeze_support

from scipy.spatial.transform import _rotation_groups
import ms_mint_app


welcome = r"""
 __________________________________________________________________________________________________________
/___/\\\\____________/\\\\__/\\\\\\\\\\\__/\\\\\_____/\\\__/\\\\\\\\\\\\\\\_______________/\\\_____________\
|___\/\\\\\\________/\\\\\\_\/////\\\///__\/\\\\\\___\/\\\_\///////\\\/////______________/\\\\\\\__________|
|____\/\\\//\\\____/\\\//\\\_____\/\\\_____\/\\\/\\\__\/\\\_______\/\\\__________________/\\\\\\\\\________|
|_____\/\\\\///\\\/\\\/_\/\\\_____\/\\\_____\/\\\//\\\_\/\\\_______\/\\\_________________\//\\\\\\\________|
|______\/\\\__\///\\\/___\/\\\_____\/\\\_____\/\\\\//\\\\/\\\_______\/\\\__________________\//\\\\\________|
|_______\/\\\____\///_____\/\\\_____\/\\\_____\/\\\_\//\\\/\\\_______\/\\\___________________\//\\\________|
|________\/\\\_____________\/\\\_____\/\\\_____\/\\\__\//\\\\\\_______\/\\\____________________\///________|
|_________\/\\\_____________\/\\\__/\\\\\\\\\\\_\/\\\___\//\\\\\_______\/\\\_____________________/\\\______|
|__________\///______________\///__\///////////__\///_____\/////________\///_____________________\///______|
\__________________________________________________________________________________________________________/
       \
        \   ^__^
         \  (@@)\_______
            (__)\       )\/\
                ||----w |
                ||     ||
"""


def _get_distribution(dist):
    if IS_FROZEN and dist == "flask-compress":
        return _Dist("1.5.0")
    else:
        return _true_get_distribution(dist)


if __name__ == "__main__":
    freeze_support()

    HOME = expanduser("~")
    DATADIR = str(P(HOME) / "MINT")
    IS_FROZEN = hasattr(sys, "_MEIPASS")

    # backup true function
    _true_get_distribution = pkg_resources.get_distribution
    # create small placeholder for the dash call
    # _flask_compress_version = parse_version(get_distribution("flask-compress").version)
    _Dist = namedtuple("_Dist", ["version"])

    # monkey patch the function so it can work once frozen and pkg_resources is of
    # no help
    pkg_resources.get_distribution = _get_distribution

    parser = argparse.ArgumentParser(description="MINT frontend.")

    parser.add_argument(
        "--no-browser",
        action="store_true",
        default=False,
        help="do not start the browser",
    )
    parser.add_argument(
        "--version", default=False, action="store_true", help="print current version"
    )
    parser.add_argument(
        "--data-dir", default=DATADIR, help="target directory for MINT data"
    )
    parser.add_argument(
        "--debug",
        default=False,
        action="store_true",
        help="start MINT server in debug mode",
    )
    parser.add_argument("--port", type=int, default=9999, help="Port to use")
    parser.add_argument(
        "--host", type=str, default="127.0.0.1", help="Host binding address"
    )
    parser.add_argument(
        "--serve-path",
        default=None,
        type=str,
        help="(deprecated) serve app at a different path e.g. '/mint/' to serve the app at 'localhost:9999/mint/'",
    )
    parser.add_argument(
        "--ncpu",
        default=None,
        type=int,
        help='Number of CPUs to use',
    )
    args = parser.parse_args()

    if args.version:
        print("Mint version:", ms_mint_app.__version__)
        exit()

    url = f"http://{args.host}:{args.port}"

    if not args.no_browser:
        if os.name == "nt":
            # https://github.com/pyinstaller/pyinstaller/wiki/Recipe-Multiprocessing
            multiprocessing.freeze_support()

        # Open the browser
        if sys.platform in ["win32", "nt"]:
            os.startfile(url)
        elif sys.platform == "darwin":
            subprocess.Popen(["open", url])
        else:
            try:
                subprocess.Popen(["xdg-open", url])
            except OSError:
                print("Please open a browser on: ", url)

    if args.data_dir is not None:
        os.environ["MINT_DATA_DIR"] = args.data_dir

    if args.serve_path is not None:
        os.environ["MINT_SERVE_PATH"] = args.serve_path

    print(welcome)

    print("Loading app...")

    from ms_mint_app.app import create_app, register_callbacks

    app, cache, fsc = create_app()
    register_callbacks(app, cache, fsc)

    app.css.config.serve_locally = True
    app.scripts.config.serve_locally = True

    print("Configuration done starting server...")

    if args.debug:

        app.run_server(
            debug=args.debug,
            port=args.port,
            host=args.host,
            dev_tools_hot_reload=False,
            dev_tools_hot_reload_interval=3000,
            dev_tools_hot_reload_max_retry=30,
        )
    else:
        serve(app.server, port=args.port, host=args.host)



import os

# import configparser

from pathlib import Path as P
from flask_login.mixins import UserMixin

from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import create_engine


db = SQLAlchemy()


class ConnectDB:
    def __init__(self, path, app):

        server = app.server

        full_path = P(path) / "data.sqlite"

        self.engine = create_engine(f"sqlite:///{full_path}")

        self.db = db

        # config = configparser.ConfigParser()

        server.config.update(
            SECRET_KEY=os.urandom(12),
            SQLALCHEMY_DATABASE_URI=f"sqlite:///{full_path}",
            SQLALCHEMY_TRACK_MODIFICATIONS=False,
        )

        app.db = self.db

        app.engine = self.engine

        db.init_app(server)

        self.create_users_table()

    def create_users_table(self):
        Users.metadata.create_all(self.engine)


class Users(db.Model, UserMixin):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(15), unique=True, nullable=False)
    email = db.Column(db.String(50), unique=True)
    password = db.Column(db.String(80))



from werkzeug.security import generate_password_hash, check_password_hash
from flask_login import login_user, logout_user, current_user

import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output, State

from sqlalchemy import Table

from .database import Users


_layout = html.Div(
    [
        html.Div(id="page-content", className="content"),
        dcc.Location(id="url", refresh=False),
    ]
)


create = html.Div(
    [
        html.H1("Create User Account"),
        dcc.Location(id="create_user", refresh=True),
        dcc.Input(id="username", type="text", placeholder="user name", maxLength=15),
        dcc.Input(id="password", type="password", placeholder="password"),
        dcc.Input(id="email", type="email", placeholder="email", maxLength=50),
        html.Button("Create User", id="submit-val", n_clicks=0),
        html.Div(id="container-button-basic"),
    ]
)  # end div


login = html.Div(
    [
        dcc.Location(id="url_login", refresh=True),
        html.H2("""Please log in to continue:""", id="h1"),
        dcc.Input(placeholder="Enter your username", type="text", id="uname-box"),
        dcc.Input(placeholder="Enter your password", type="password", id="pwd-box"),
        html.Button(children="Login", n_clicks=0, type="submit", id="login-button"),
        html.Div(children="", id="output-state"),
    ]
)  # end div


success = html.Div(
    [
        dcc.Location(id="url_login_success", refresh=True),
        html.Div(
            [
                html.Br(),
                html.Button(
                    id="back-button",
                    children="Logout",
                    n_clicks=0,
                    style={"float": "right"},
                ),
            ]
        ),
    ]
)

failed = html.Div(
    [
        dcc.Location(id="url_login_df", refresh=True),
        html.Div(
            [
                html.H2("Log in Failed. Please try again."),
                html.Br(),
                html.Div([login]),
                html.Br(),
                html.Button(id="back-button", children="Logout", n_clicks=0),
            ]
        ),  # end div
    ]
)  # end div


logout = html.Div(
    [
        dcc.Location(id="logout", refresh=True),
        html.Br(),
        html.Div(html.H2("You have been logged out - Please login")),
        html.Br(),
        html.Div([login]),
        html.Button(id="back-button", children="Logout", n_clicks=0),
    ]
)  # end div


def callbacks(app, fsc=None, cache=None):
    @app.callback(
        Output("container-button-basic", "children"),
        Input("submit-val", "n_clicks"),
        State("username", "value"),
        State("password", "value"),
        State("email", "value"),
    )
    def insert_users(n_clicks, un, pw, em):
        hashed_password = generate_password_hash(pw, method="sha256")
        if un is not None and pw is not None and em is not None:
            Users_tbl = Table("users", Users.metadata)
            ins = Users_tbl.insert().values(
                username=un,
                password=hashed_password,
                email=em,
            )
            with app.engine.begin() as conn:
                conn.execute(ins)
            return [login]
        else:
            return [
                html.Div(
                    [
                        html.H2("Already have a user account?"),
                        dcc.Link("Click here to Log In", href="/login"),
                    ]
                )
            ]

    @app.callback(
        Output("url_login_success", "pathname"), [Input("back-button", "n_clicks")]
    )
    def logout_dashboard(n_clicks):
        if n_clicks > 0:
            return "/"

    @app.callback(Output("page-content", "children"), [Input("url", "pathname")])
    def display_page(pathname):
        if pathname == "/":
            return create
        elif pathname == "/login":
            return login
        elif pathname == "/success":
            if current_user.is_authenticated:
                return success
            else:
                return failed
        elif pathname == "/data":
            if current_user.is_authenticated:
                return None
        elif pathname == "/logout":
            if current_user.is_authenticated:
                logout_user()
                return logout
            else:
                return logout
        else:
            return "404"

    @app.callback(
        Output("url_login", "pathname"),
        Input("login-button", "n_clicks"),
        State("uname-box", "value"),
        State("pwd-box", "value"),
    )
    def successful(n_clicks, input1, input2):
        user = Users.query.filter_by(username=input1).first()
        if user:
            if check_password_hash(user.password, input2):
                login_user(user)
                return "/success"
            else:
                pass
        else:
            pass

    @app.callback(
        Output("output-state", "children"),
        Input("login-button", "n_clicks"),
        State("uname-box", "value"),
        State("pwd-box", "value"),
    )
    def update_output(n_clicks, input1, input2):
        if n_clicks > 0:

            user = Users.query.filter_by(username=input1).first()

            if user:
                if check_password_hash(user.password, input2):
                    return ""
                else:
                    return "Incorrect username or password"
            else:
                return "Incorrect username or password"
        else:
            return ""



import os
import tempfile
import logging

import pandas as pd

from pathlib import Path as P

import matplotlib

matplotlib.use("Agg")

import dash
from dash import html, dcc

from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
from dash.dcc import Download
from dash_extensions.enrich import FileSystemCache

import dash_bootstrap_components as dbc

from flask_caching import Cache
from flask_login import current_user


from . import tools as T

from . import workspaces
from . import ms_files
from . import metadata
from . import targets
from . import peak_optimization
from . import processing
#from . import add_metab
from . import analysis
from . import messages

import dash_uploader as du


def make_dirs():
    tmpdir = tempfile.gettempdir()
    tmpdir = os.path.join(tmpdir, "MINT")
    tmpdir = os.getenv("MINT_DATA_DIR", default=tmpdir)
    cachedir = os.path.join(tmpdir, ".cache")
    os.makedirs(tmpdir, exist_ok=True)
    os.makedirs(cachedir, exist_ok=True)
    print("MAKEDIRS:", tmpdir, cachedir)
    return P(tmpdir), P(cachedir)


TMPDIR, CACHEDIR = make_dirs()

config = {
    "DEBUG": True,  # some Flask specific configs
    "CACHE_TYPE": "simple",  # Flask-Caching related configs
    "CACHE_DEFAULT_TIMEOUT": 300,
}

pd.options.display.max_colwidth = 1000

_modules = [
    workspaces,
    ms_files,
    metadata,
    targets,
    #add_metab,
    peak_optimization,
    processing,
    analysis,
]

modules = {module._label: module for module in _modules}

# Collect outputs:
_outputs = html.Div(
    id="outputs",
    children=[module._outputs for module in _modules if module._outputs is not None],
    style={"visibility": "hidden"},
)


logout_button = (
    dbc.Button(
        "Logout",
        id="logout-button",
        style={"marginRight": "10px", "visibility": "hidden"},
    ),
)
logout_button = html.A(href="/logout", children=logout_button)

_layout = html.Div(
    [
        html.Div(logout_button),
        dcc.Interval(
            id="progress-interval", n_intervals=0, interval=2000, disabled=False
        ),
        html.A(
            href="https://lewisresearchgroup.github.io/ms-mint-app/gui/",
            children=[
                html.Button(
                    "Documentation",
                    id="B_help",
                    style={"float": "right", "color": "info"},
                )
            ],
            target="_blank",
        ),
        html.A(
            href=f"https://github.com/LewisResearchGroup/ms-mint-app/issues/new?body={T.get_issue_text()}",
            children=[
                html.Button(
                    "Issues", id="B_issues", style={"float": "right", "color": "info"}
                )
            ],
            target="_blank",
        ),
        dbc.Progress(
            id="progress-bar",
            value=100,
            style={"marginBottom": "20px", "width": "100%", "marginTop": "20px"},
        ),
        messages.layout(),
        Download(id="res-download-data"),
        html.Div(id="tmpdir", children=str(TMPDIR), style={"visibility": "hidden"}),
        html.Div(
            [
                html.P(
                    "Current Workspace: ",
                    style={
                        "display": "inline-block",
                        "marginRight": "5px",
                        "marginTop": "5px",
                    },
                ),
                html.Div(id="active-workspace", style={"display": "inline-block"}),
                html.Div(
                    id="wdir",
                    children="",
                    style={
                        "display": "inline-block",
                        "visibility": "visible",
                        "float": "right",
                    },
                ),
            ],
            style={"marginBottom": "20px"},
        ),
        html.Div(id="pko-creating-chromatograms"),
        dcc.Tabs(
            id="tab",
            value=_modules[0]._label,
            children=[
                dcc.Tab(
                    id=modules[key]._label,
                    value=key,
                    label=modules[key]._label,
                )
                for key in modules.keys()
            ],
        ),
        html.Div(id="pko-image-store", style={"visibility": "hidden", "height": "0px"}),
        html.Div(id="tab-content"),
        html.Div(id="viewport-container", style={"visibility": "hidden"}),
        _outputs,
    ],
    style={"margin": "2%"},
)


def register_callbacks(app, cache, fsc):
    logging.warning("Register callbacks")
    upload_root = os.getenv("MINT_DATA_DIR", tempfile.gettempdir())
    upload_dir = str(P(upload_root) / "MINT-Uploads")
    UPLOAD_FOLDER_ROOT = upload_dir
    du.configure_upload(app, UPLOAD_FOLDER_ROOT)

    messages.callbacks(app=app, fsc=fsc, cache=cache)

    for module in _modules:
        func = module.callbacks
        if func is not None:
            func(app=app, fsc=fsc, cache=cache)

    # Updates the current viewport
    app.clientside_callback(
        """
        function(href) {
            var w = window.innerWidth;
            var h = window.innerHeight;
            return `${w},${h}` ;
        }
        """,
        Output("viewport-container", "children"),
        Input("progress-interval", "n_intervals"),
    )

    @app.callback(
        Output("tab-content", "children"),
        Input("tab", "value"),
        State("wdir", "children"),
    )
    def render_content(tab, wdir):
        func = modules[tab].layout
        if tab != "Workspaces" and wdir == "":
            return dbc.Alert(
                "Please, create and activate a workspace.", color="warning"
            )
        elif (
            tab in ["Metadata", "Peak Optimization", "Processing"]
            and len(T.get_ms_fns(wdir)) == 0
        ):
            return dbc.Alert("Please import MS files.", color="warning")
        elif tab in ["Processing"] and (len(T.get_targets(wdir)) == 0):
            return dbc.Alert("Please, define targets.", color="warning")
        elif tab in ["Analysis"] and not P(T.get_results_fn(wdir)).is_file():
            return dbc.Alert("Please, create results (Processing).", color="warning")
        if func is not None:
            return func()
        else:
            raise PreventUpdate

    @app.callback(
        Output("tmpdir", "children"),
        Output("logout-button", "style"),
        Input("progress-interval", "n_intervals"),
    )
    def upate_tmpdir(x):
        if hasattr(app.server, "login_manager"):
            username = current_user.username
            logging.warning(f"User: {username}")
            tmpdir = str(TMPDIR / "User" / username)
            logging.warning(tmpdir)
            return tmpdir, {"visibility": "visible"}
        logging.info("Hide login button")
        return str(TMPDIR / "Local"), {"visibility": "hidden"}


def create_app(**kwargs):

    app = dash.Dash(
        __name__,
        external_stylesheets=[
            dbc.themes.MINTY,
            "https://codepen.io/chriddyp/pen/bWLwgP.css",
        ],
        **kwargs,
    )

    app.layout = _layout
    app.title = "MINT"
    app.config["suppress_callback_exceptions"] = True

    upload_root = os.getenv("MINT_DATA_DIR", tempfile.gettempdir())
    CACHE_DIR = str(P(upload_root) / "MINT-Cache")

    logging.info("Cache directory: {}".format(CACHE_DIR))

    cache = Cache(
        app.server, config={"CACHE_TYPE": "filesystem", "CACHE_DIR": CACHE_DIR}
    )

    fsc = FileSystemCache(str(CACHEDIR))

    return app, cache, fsc


if __name__ == "__main__":
    app, cache = create_app()
    register_callbacks(app, cache)
    app.run_server(
        debug=True,
        threaded=True,
        dev_tools_hot_reload_interval=5000,
        dev_tools_hot_reload_max_retry=30,
    )



import os
import shutil
import uuid
import logging
import numpy as np
import tempfile

from pathlib import Path as P

import pandas as pd

from dash import html, dcc
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State

import dash_bootstrap_components as dbc

from ms_mint.io import convert_ms_file_to_feather

from dash_tabulator import DashTabulator

import dash_uploader as du

from . import tools as T


upload_root = os.getenv("MINT_DATA_DIR", tempfile.gettempdir())
upload_dir = str(P(upload_root) / "MINT-Uploads")
UPLOAD_FOLDER_ROOT = upload_dir


options = {
    "selectable": True,
    "headerFilterLiveFilterDelay": 3000,
    "layout": "fitDataFill",
    "height": "900px",
}

clearFilterButtonType = {"css": "btn btn-outline-dark", "text": "Clear Filters"}


columns = [
    {
        "formatter": "rowSelection",
        "titleFormatter": "rowSelection",
        "titleFormatterParams": {
            "rowRange": "active"  # only toggle the values of the active filtered rows
        },
        "hozAlign": "center",
        "headerSort": False,
        "width": "1px",
        "frozen": True,
    },
    {
        "title": "MS-file",
        "field": "MS-file",
        "headerFilter": True,
        "headerSort": True,
        "editor": None,
        "width": "80%",
        "sorter": "string",
        "frozen": True,
    },
    {
        "title": "Size [MB]",
        "field": "file_size",
        "headerFilter": True,
        "headerSort": True,
        "editor": None,
        "width": "20%",
        "sorter": "string",
        "frozen": True,
    },
    {
        "title": "",
        "field": "",
        "headerFilter": False,
        "formatter": "color",
        "width": "3px",
        "headerSort": False,
    },
]


ms_table = html.Div(
    id="ms-table-container",
    style={"Height": 0, "marginTop": "10%"},
    children=[
        DashTabulator(
            id="ms-table",
            columns=columns,
            options=options,
            clearFilterButtonType=clearFilterButtonType,
        )
    ],
)

_label = "MS-Files"

_layout = html.Div(
    [
        html.H3("Upload MS-files"),
        html.Div(
            du.Upload(
                id="ms-uploader",
                filetypes=["tar", "zip", "mzxml", "mzml", "mzXML", "mzML"],
                upload_id=uuid.uuid1(),
                max_files=10000,
                pause_button=True,
                cancel_button=True,
                text="Upload mzXML/mzML files.",
            ),
            style={
                "textAlign": "center",
                "width": "100%",
                "padding": "0px",
                "margin-bottom": "20px",
                "display": "inline-block",
            },
        ),
        dcc.Markdown("---", style={"marginTop": "10px"}),
        dcc.Markdown("##### Actions"),
        html.Button("Convert to Feather", id="ms-convert"),
        html.Button("Delete selected files", id="ms-delete", style={"float": "right"}),
        html.Div(id="ms-n-files"),
        dcc.Loading(ms_table),
        html.Div(id="ms-uploader-fns", style={"visibility": "hidden"}),
    ]
)


_outputs = html.Div(
    id="ms-outputs",
    children=[
        html.Div(id={"index": "ms-convert-output", "type": "output"}),
        html.Div(id={"index": "ms-delete-output", "type": "output"}),
        html.Div(id={"index": "ms-save-output", "type": "output"}),
        html.Div(id={"index": "ms-import-from-url-output", "type": "output"}),
        html.Div(id={"index": "ms-uploader-output", "type": "output"}),
        html.Div(id={"index": "ms-new-target-output", "type": "output"}),
    ],
)


def layout():
    return _layout


def callbacks(app, fsc, cache):
    @app.callback(
        Output("ms-table", "data"),
        Output("ms-table", "downloadButtonType"),
        Input({"index": "ms-uploader-output", "type": "output"}, "children"),
        Input("wdir", "children"),
        Input({"index": "ms-delete-output", "type": "output"}, "children"),
        Input({"index": "ms-convert-output", "type": "output"}, "children"),
        State("active-workspace", "children"),
    )
    def ms_table(value, wdir, files_deleted, files_converted, workspace):

        ms_files = T.get_ms_fns(wdir)

        data = pd.DataFrame(
            {
                "MS-file": [os.path.basename(fn) for fn in ms_files],
                "file_size": [
                    np.round(os.path.getsize(fn) / 1024 / 1024, 2) for fn in ms_files
                ],
            }
        )

        downloadButtonType = {
            "css": "btn btn-primary",
            "text": "Export",
            "type": "csv",
            "filename": f"{T.today()}__MINT__{workspace}__ms-files",
        }

        return data.to_dict("records"), downloadButtonType

    @app.callback(
        Output({"index": "ms-convert-output", "type": "output"}, "children"),
        Input("ms-convert", "n_clicks"),
        State("ms-table", "multiRowsClicked"),
        State("wdir", "children"),
    )
    def ms_convert(n_clicks, rows, wdir):
        target_dir = os.path.join(wdir, "ms_files")
        if n_clicks is None:
            raise PreventUpdate
        fns = [row["MS-file"] for row in rows]
        fns = [fn for fn in fns if not fn.endswith(".feather")]
        fns = [os.path.join(target_dir, fn) for fn in fns]
        n_total = len(fns)
        for i, fn in enumerate(fns):
            fsc.set("progress", int(100 * (i + 1) / n_total))
            new_fn = convert_ms_file_to_feather(fn)
            if os.path.isfile(new_fn):
                os.remove(fn)
        return dbc.Alert("Files converted to feather format.", color="info")

    @app.callback(
        Output({"index": "ms-delete-output", "type": "output"}, "children"),
        Input("ms-delete", "n_clicks"),
        State("ms-table", "multiRowsClicked"),
        State("wdir", "children"),
    )
    def ms_delete(n_clicks, rows, wdir):
        if n_clicks is None:
            raise PreventUpdate
        target_dir = os.path.join(wdir, "ms_files")
        for row in rows:
            fn = row["MS-file"]
            fn = P(target_dir) / fn
            os.remove(fn)
        return dbc.Alert(f"{len(rows)} files deleted", color="info")


    @du.callback(
        output=Output('ms-uploader-fns', 'children'),
        id='ms-uploader',
    )
    def upload_completed(status):
        return [str(fn) for fn in status.uploaded_files]

    @app.callback(
        Output({"index": "ms-uploader-output", "type": "output"}, "children"),
        Input("ms-uploader-fns", "children"),
        State("wdir", "children"),
    )
    def move_uploaded_files(fns, wdir):
        if fns is None or len(fns)==0:
            raise PreventUpdate
        ms_dir = T.get_ms_dirname(wdir)
        for fn in fns:
            if not P(fn).is_file():
                continue
            fn_new = P(ms_dir) / P(fn).name
            shutil.move(fn, fn_new)
            logging.info(f"Move {fn} to {fn_new}")
        return dbc.Alert("Upload finished", color="success")

    @app.callback(Output("ms-n-files", "children"), Input("ms-table", "data"))
    def n_files(data):
        n_files = len(data)
        return dbc.Alert(f"{n_files} files in current workspace.", color="success")




import re
import shutil

import pandas as pd

import dash
from dash import html, dcc, dash_table
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State
import dash_bootstrap_components as dbc


from . import tools as T


ws_table = html.Div(
    id="ws-table-container",
    style={"minHeight": 100, "margin": "5%"},
    children=[
        dash_table.DataTable(
            id="ws-table",
            columns=[{"name": i, "id": i, "selectable": True} for i in ["Workspace"]],
            data=[],
            row_selectable="single",
            row_deletable=False,
            style_cell={"textAlign": "left"},
            sort_action="native",
        )
    ],
)


_layout = html.Div(
    [
        html.H3("Workspaces"),
        html.Button("Create Workspace", id="ws-create"),
        html.Button("Delete Workspace", id="ws-delete", style={"float": "right"}),
        ws_table,
        dbc.Modal(
            [
                dbc.ModalHeader("Create Workspace"),
                dbc.ModalBody(
                    [
                        dcc.Input(
                            id="ws-create-input", placeholder="New workspace name"
                        ),
                        html.P(id="ws-create-output"),
                    ]
                ),
                dbc.ModalFooter(
                    html.Div(
                        [
                            dbc.Button(
                                "Create",
                                id="ws-create-confirm",
                                style={"marginRight": "10px"},
                            ),
                            dbc.Button("Cancel", id="ws-create-cancel"),
                        ]
                    )
                ),
            ],
            id="ws-create-popup",
        ),
        dbc.Modal(
            [
                dbc.ModalHeader("Delete Workspace"),
                dbc.ModalBody(
                    "This will delete all files and results in the selected workspace."
                ),
                dbc.ModalFooter(
                    html.Div(
                        [
                            dbc.Button(
                                "Delete",
                                id="ws-delete-confirm",
                                style={"marginRight": "10px"},
                            ),
                            dbc.Button("Cancel", id="ws-delete-cancel"),
                        ]
                    )
                ),
            ],
            id="ws-delete-popup",
        ),
    ]
)


_outputs = html.Div(
    id="ws-outputs",
    children=[
        html.Div(id={"index": "ws-created-output", "type": "output"}),
        html.Div(id={"index": "ws-activate-output", "type": "output"}),
        html.Div(id={"index": "ws-delete-output", "type": "output"}),
    ],
)

_label = "Workspaces"


def layout():
    return _layout


def callbacks(app, fsc, cache):
    @app.callback(
        Output("ws-table", "data"),
        Input({"index": "ws-created-output", "type": "output"}, "children"),
        Input("tab", "value"),
        Input({"index": "ws-delete-output", "type": "output"}, "children"),
        State("tmpdir", "children"),
    )
    def ws_table(value, tab, delete, tmpdir):
        T.maybe_migrate_workspaces(tmpdir)
        ws_names = T.get_workspaces(tmpdir)
        ws_names.sort()
        ws_names = [
            {"Workspace": ws_name}
            for ws_name in ws_names
            if not ws_name.startswith(".")
        ]
        return ws_names

    @app.callback(
        Output({"index": "ws-activate-output", "type": "output"}, "children"),
        Output("wdir", "children"),
        Output("active-workspace", "children"),
        Input("ws-table", "derived_virtual_selected_rows"),
        Input({"index": "ws-delete-output", "type": "output"}, "children"),
        Input({"index": "ws-created-output", "type": "output"}, "children"),
        State("ws-table", "data"),
        State("tmpdir", "children"),
    )
    def ws_activate(ndx, deleted, created, data, tmpdir):
        prop_id = dash.callback_context.triggered[0]["prop_id"]
        if tmpdir is None:
            raise PreventUpdate
        ws_names = T.get_workspaces(tmpdir)
        if ws_names is None or len(ws_names) == 0:
            message = "No workspace defined."
            return dbc.Alert(message, color="danger"), "", ""
        if prop_id == "ws-delete-output.children" or ndx is None:
            ndx = [0]

        data = pd.DataFrame(data)

        if len(ndx) == 1:
            if ndx[0] not in data.index:
                raise PreventUpdate
            ws_name = data.loc[ndx[0], "Workspace"]
        else:
            ws_name = T.get_active_workspace(tmpdir)

        if ws_name is None:
            raise PreventUpdate

        wdir = T.workspace_path(tmpdir, ws_name)

        if ws_name is not None:
            T.save_activated_workspace(tmpdir, ws_name)
        else:
            raise PreventUpdate
        message = f"Workspace {ws_name} activated."
        if ws_name is None:
            ws_name = ""

        T.maybe_update_workpace_scheme(wdir)

        return dbc.Alert(message, color="info"), wdir, ws_name

    @app.callback(
        Output("progress-bar", "value"),
        Input("progress-interval", "n_intervals"),
    )
    def set_progress(n):
        return fsc.get("progress")

    @app.callback(
        Output("ws-delete-popup", "is_open"),
        Input("ws-delete", "n_clicks"),
        Input("ws-delete-cancel", "n_clicks"),
        Input("ws-delete-confirm", "n_clicks"),
        State("ws-delete-popup", "is_open"),
    )
    def ws_delete(n1, n2, n3, is_open):
        if n1 is None:
            raise PreventUpdate
        if n1 or n2 or n3:
            return not is_open
        return is_open

    @app.callback(
        Output({"index": "ws-delete-output", "type": "output"}, "children"),
        Input("ws-delete-confirm", "n_clicks"),
        State("ws-table", "derived_virtual_selected_rows"),
        State("ws-table", "data"),
        State("tmpdir", "children"),
    )
    def ws_delete_confirmed(n_clicks, ndxs, data, tmpdir):
        if n_clicks is None or len(ndxs) == 0:
            raise PreventUpdate
        for ndx in ndxs:
            ws_name = data[ndx]["Workspace"]
            dirname = T.workspace_path(tmpdir, ws_name)
            shutil.rmtree(dirname)
        message = f"Worskpace {ws_name} deleted."
        return dbc.Alert(message, color="info")

    @app.callback(
        Output("ws-create-popup", "is_open"),
        Input("ws-create", "n_clicks"),
        Input("ws-create-cancel", "n_clicks"),
        Input("ws-create-confirm", "n_clicks"),
        State("ws-create-popup", "is_open"),
    )
    def ws_create(n1, n2, n3, is_open):
        if n1 is None:
            raise PreventUpdate
        if n1 or n2 or n3:
            return not is_open
        return is_open

    @app.callback(
        Output("ws-create-output", "children"),
        Output("ws-create-confirm", "disabled"),
        Input("ws-create-input", "value"),
        State("tmpdir", "children"),
    )
    def ws_create_message(ws_name, tmpdir):
        if ws_name is None:
            raise PreventUpdate
        elif ws_name == "":
            return dbc.Alert("Name cannot be empty", color="warning"), True
        elif not re.match("^[\w_-]+$", ws_name):
            return (
                dbc.Alert(
                    "Name can only contain: a-z, A-Z, 0-9, -,  _ and no blanks.",
                    color="warning",
                ),
                True,
            )
        elif T.workspace_exists(tmpdir, ws_name):
            return dbc.Alert("Workspace already exists", color="warning"), True
        else:
            return (
                dbc.Alert(f'Can create workspace "{ws_name}"', color="success"),
                False,
            )

    @app.callback(
        Output({"index": "ws-created-output", "type": "output"}, "children"),
        Input("ws-create-confirm", "n_clicks"),
        State("ws-create-input", "value"),
        State("tmpdir", "children"),
    )
    def ws_create_confirmed(n_clicks, ws_name, tmpdir):
        if n_clicks is None:
            raise PreventUpdate
        T.create_workspace(tmpdir, ws_name)
        return dbc.Alert(f"Workspace {ws_name} created.", color="success")

    @app.callback(
        Output("ws-table", "selected_rows"),
        Input("tab", "value"),
        Input("ws-table", "data"),
        State("tmpdir", "children"),
    )
    def set_selected_row(tab, data, tmpdir):
        if tab != _label:
            raise PreventUpdate

        ws_names = T.get_workspaces(tmpdir)

        if len(ws_names) == 0:
            raise PreventUpdate
        data = pd.DataFrame(data)
        active_ws = T.get_active_workspace(tmpdir)
        if active_ws is None:
            ndx = data.index[0]
            active_ws = data.Workspace[0]
        else:
            ndx = ws_names.index(active_ws)
        if ndx is None:
            ndx = 0
        return [ndx]



import os

from pathlib import Path as P

import dash
from dash import html

import dash_bootstrap_components as dbc
from dash.exceptions import PreventUpdate
from dash.dcc import send_file, send_bytes

from dash.dependencies import Input, Output, State

from ms_mint.Mint import Mint

from . import tools as T

_label = "Processing"


_layout = html.Div(
    [
        html.H3("Run MINT"),
        html.Button("Run MINT", id="run-mint"),
        html.Button("Download all results", id="res-download"),
        html.Button("Download dense peak_max", id="res-download-peakmax"),
        html.Button("Delete results", id="res-delete", style={"float": "right"}),
    ]
)

_outputs = html.Div(
    id="run-outputs",
    children=[
        html.Div(
            id={"index": "run-mint-output", "type": "output"},
            style={"visibility": "hidden"},
        ),
    ],
)


def layout():
    return _layout


def callbacks(app, fsc, cache):
    @app.callback(
        Output("res-delete-output", "children"),
        Input("res-delete", "n_clicks"),
        State("wdir", "children"),
    )
    def heat_delete(n_clicks, wdir):
        if n_clicks is None:
            raise PreventUpdate
        os.remove(T.get_results_fn(wdir))
        return dbc.Alert("Results file deleted.")

    @app.callback(
        [
            Output("res-download-data", "data"),
            Input("res-download", "n_clicks"),
            Input("res-download-peakmax", "n_clicks"),
            State("wdir", "children"),
        ]
    )
    def download_results(n_clicks, n_clicks_peakmax, wdir):
        if (n_clicks is None) and (n_clicks_peakmax is None):
            raise PreventUpdate
        ctx = dash.callback_context

        prop_id = ctx.triggered[0]["prop_id"]

        if prop_id == "res-download.n_clicks":
            fn = T.get_results_fn(wdir)
            workspace = os.path.basename(wdir)
            return [
                send_file(fn, filename=f"{T.today()}__MINT__{workspace}__results.csv")
            ]

        elif prop_id == "res-download-peakmax.n_clicks":
            workspace = os.path.basename(wdir)
            results = T.get_results(wdir)
            df = results.pivot_table("peak_max", "peak_label", "ms_file")
            df.columns = [P(x).with_suffix("") for x in df.columns]
            buffer = T.df_to_in_memory_excel_file(df)
            return [
                send_bytes(
                    buffer, filename=f"{T.today()}__MINT__{workspace}__peak-max.xlsx"
                )
            ]

    @app.callback(
        Output({"index": "run-mint-output", "type": "output"}, "children"),
        Input("run-mint", "n_clicks"),
        State("wdir", "children"),
    )
    def run_mint(n_clicks, wdir):
        if n_clicks is None:
            raise PreventUpdate

        def set_progress(x):
            fsc.set("progress", x)

        mint = Mint(verbose=False, progress_callback=set_progress)
        targets_fn = T.get_targets_fn(wdir)
        output_fn = T.get_results_fn(wdir)
        try:
            mint.load_targets(targets_fn)
            mint.ms_files = T.get_ms_fns(wdir)
            mint.run(output_fn=output_fn)
        except Exception as e:
            return dbc.Alert(str(e), color="danger")
        return dbc.Alert("Finished running MINT", color="success")



import os
import pandas as pd

from dash import html, dcc
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
from dash_tabulator import DashTabulator

import dash_bootstrap_components as dbc

from ms_mint.standards import M_PROTON
from . import tools as T

from pkg_resources import resource_filename

CHEBI_CHEM_FN = os.path.abspath(
    resource_filename("ms_mint_app.static", "ChEBI-Chem.parquet")
)
CHEBI_GROUPS_FN = os.path.abspath(
    resource_filename("ms_mint_app.static", "ChEBI-Groups.parquet")
)
CHEBI_CHEM = pd.read_parquet(CHEBI_CHEM_FN)
CHEBI_GROUPS = pd.read_parquet(CHEBI_GROUPS_FN)


options = {
    "selectable": True,
    "headerFilterLiveFilterDelay": 3000,
    "layout": "fitDataFill",
    "height": "600px",
}

clearFilterButtonType = {"css": "btn btn-outline-dark", "text": "Clear Filters"}


groups_options = [{"label": "All", "value": "all"}] + [
    {"label": x.capitalize(), "value": x} for x in CHEBI_GROUPS.index
]

columns = T.gen_tabulator_columns(CHEBI_CHEM.columns, editor=None, col_width="auto")

add_metab_table = html.Div(
    id="add-metab-table-container",
    style={"minHeight": 100, "marginTop": "10%"},
    children=[
        dbc.Alert(
            "This feature is experimental. Please, check values and report errors to the issue tracker.",
            color="danger",
        ),
        dcc.Dropdown(
            id="add-metab-ms-mode",
            placeholder="Select ionization mode",
            options=[
                {"label": "Positive", "value": "Positive"},
                {"label": "Negative", "value": "Negative"},
            ],
            value="Negative",
        ),
        dcc.Dropdown(id="add-metab-groups", options=groups_options, multi=True),
        dcc.Loading(
            DashTabulator(
                id="add-metab-table",
                columns=columns,
                options=options,
                clearFilterButtonType=clearFilterButtonType,
            ),
        ),
        html.Button("Add selected metabolites to targets", id="add-metab"),
    ],
)

_label = "Add Metabolites"

_layout = html.Div([html.H3("Add Metabolites"), add_metab_table])

_outputs = html.Div(
    id="add-metab-outputs",
    children=[html.Div(id={"index": "add-metab-output", "type": "output"})],
)


def layout():
    return _layout


def callbacks(app, fsc, cache):
    @app.callback(
        Output({"index": "add-metab-output", "type": "output"}, "children"),
        Input("add-metab", "n_clicks"),
        State("add-metab-table", "multiRowsClicked"),
        State("add-metab-ms-mode", "value"),
        State("wdir", "children"),
    )
    def add_metab(n_clicks, rows, ms_mode, wdir):
        if n_clicks is None:
            raise PreventUpdate
        if ms_mode is None:
            return dbc.Alert("Please select ionization mode.", color="warning")

        targets = T.get_targets(wdir)

        for row in rows:
            charge = int(row["Charge"])
            if (ms_mode == "Negative") and (charge > 0):
                continue
            if (ms_mode == "Positive") and (charge < 0):
                continue
            if charge == 0:
                if ms_mode == "Negative":
                    targets.loc[row["ChEBI Name"], "mz_mean"] = (
                        row["Monoisotopic Mass"] - M_PROTON
                    )
                elif ms_mode == "Positive":
                    targets.loc[row["ChEBI Name"], "mz_mean"] = row[
                        "Monoisotopic Mass"
                    ]  # + M_PROTON
            else:
                targets.loc[row["ChEBI Name"], "mz_mean"] = row["Monoisotopic Mass"]

            targets.loc[row["ChEBI Name"], "mz_width"] = 10
            targets.loc[row["ChEBI Name"], "rt"] = -1
            targets.loc[row["ChEBI Name"], "rt_min"] = 0
            targets.loc[row["ChEBI Name"], "rt_max"] = 15
            targets.loc[row["ChEBI Name"], "intensity_threshold"] = 0

        T.write_targets(targets, wdir)
        n_peaks = len(targets)
        n_new = len(rows)
        return dbc.Alert(
            f"{n_new} peaks added, now {n_peaks} peaks defined.", color="info"
        )

    @app.callback(
        Output("add-metab-table", "data"),
        Input("add-metab-groups", "value"),
        Input("add-metab-ms-mode", "value"),
    )
    def generate_table(groups, ms_mode):
        print("Groups to add", groups)

        data = CHEBI_CHEM

        if ms_mode == "Positive":
            data = data[data.Charge.str.startswith("+") | (data.Charge == "0")]
        elif ms_mode == "Negative":
            data = data[data.Charge.str.startswith("-") | (data.Charge == "0")]

        if groups is None or len(groups) == 0:
            raise PreventUpdate
        elif "all" in groups:
            return data.to_dict("records")

        ids = (
            CHEBI_GROUPS.loc[groups]
            .explode("ChEBI IDs")
            .drop_duplicates()["ChEBI IDs"]
            .values
        )

        return data[data["ChEBI ID"].isin(ids)].to_dict("records")



from setuptools import setup, find_packages

import versioneer


with open("README.md", "r") as fh:
    long_description = fh.read()


install_requires = [
    "ms-mint==0.1.7",
    "xlsxwriter",
    "waitress",
    "dash",
    "dash_extensions",
    "dash_bootstrap_components",
    "flask_login",
    "urllib3",
    "dash_tabulator",
    "dash_uploader==0.7.0a1",
]


config = {
    "name": "ms-mint-app",
    "version": versioneer.get_version(),
    "cmdclass": versioneer.get_cmdclass(),
    "description": "Metabolomics Integrator (Mint)",
    "long_description": long_description,
    "long_description_content_type": "text/markdown",
    "author": "Soren Wacker",
    "url": "https://github.com/LewisResearchGroup/ms-mint-app",
    "author_email": "swacker@ucalgary.ca",
    "scripts": ["scripts/Mint.py"],
    "packages": find_packages(),
    "data_files": [("scripts", ["scripts/Mint.py"])],
    "classifiers": [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    "python_requires": ">=3.8",
    "install_requires": install_requires,
    "include_package_data": True,
    "package_data": {
        "ms_mint_app.static": [
            "Standard_Peaklist.csv",
            "ChEBI-Chem.parquet",
            "ChEBI-Groups.parquet",
        ]
    },
}

setup(**config)



from dash import html, dcc
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc

from . import heatmap
from . import pca
from . import distributions
from . import hierachical_clustering
from . import plotting
from . import tools as T


_modules = [heatmap, distributions, pca, hierachical_clustering, plotting]

modules = {module._label: module for module in _modules}

groupby_options = [
    {"label": "Batch", "value": "Batch"},
    {"label": "Label", "value": "Label"},
    {"label": "Type", "value": "Type"},
    {"label": "Color", "value": "Color"},
]

ana_normalization_cols = [
    {"label": "Batch", "value": "Batch"},
    {"label": "peak_label", "value": "peak_label"},
    {"label": "ms_file", "value": "ms_file"},
]

_layout = html.Div(
    [
        dcc.Tabs(
            id="ana-secondary-tab",
            value=_modules[0]._label,
            vertical=False,
            children=[
                dcc.Tab(
                    value=key,
                    label=modules[key]._label,
                )
                for key in modules.keys()
            ],
        ),
        dbc.Row([
            dbc.Col([
                dcc.Dropdown(
                    id="ana-file-types",
                    options=[],
                    placeholder="Types of files to include",
                    multi=True,
                ),
                dcc.Dropdown(
                    id="ana-peak-labels-include",
                    options=[],
                    placeholder="Include peak_labels",
                    multi=True,
                ),
                dcc.Dropdown(
                    id="ana-peak-labels-exclude",
                    options=[],
                    placeholder="Exclude peak_labels",
                    multi=True,
                ),
            ]),
            dbc.Col([
                dcc.Dropdown(
                    id="ana-ms-order", options=[], placeholder="MS-file sorting", multi=True
                ),
                dcc.Dropdown(
                    id="ana-groupby",
                    options=groupby_options,
                    value=None,
                    placeholder="Group by column",
                ),
                dcc.Dropdown(
                    id="ana-normalization-cols",
                    options=ana_normalization_cols,
                    value=None,
                    placeholder="Normalize by",
                    multi=True,
                ),
            ]),
        ]),

        html.Div(id="ana-secondary-tab-content"),
    ]
)


_label = "Analysis"

_outputs = None


def layout():
    return _layout


def callbacks(app, fsc, cache):

    for module in _modules:
        func = module.callbacks
        if func is not None:
            func(app=app, fsc=fsc, cache=cache)

    @app.callback(
        Output("ana-secondary-tab-content", "children"),
        Input("ana-secondary-tab", "value"),
        State("wdir", "children"),
    )
    def render_content(tab, wdir):
        func = modules[tab].layout
        if func is not None:
            return func()
        else:
            raise PreventUpdate

    @app.callback(
        Output("ana-file-types", "options"),
        Output("ana-file-types", "value"),
        Input("tab", "value"),
        State("wdir", "children"),
    )
    def file_types(tab, wdir):
        if tab != _label:
            raise PreventUpdate
        meta = T.get_metadata(wdir)
        if meta is None:
            raise PreventUpdate
        file_types = meta["Type"].drop_duplicates().sort_values()
        options = [{"value": str(i), "label": str(i)} for i in file_types]
        print(file_types, options)
        return options, file_types

    @app.callback(
        Output("ana-ms-order", "options"),
        Output("ana-groupby", "options"),
        Input("ana-secondary-tab", "value"),
        State("wdir", "children"),
    )
    def ms_order_options(tab, wdir):
        cols = T.get_metadata(wdir).dropna(how="all", axis=1).columns.to_list()
        if "index" in cols:
            cols.remove("index")
        if "PeakOpt" in cols:
            cols.remove("PeakOpt")
        options = [{"value": i, "label": i} for i in cols]
        return options, options

    @app.callback(
        Output("ana-peak-labels-include", "options"),
        Output("ana-peak-labels-exclude", "options"),
        Input("tab", "value"),
        State("wdir", "children"),
    )
    def peak_labels(tab, wdir):
        if tab != _label:
            raise PreventUpdate
        peaklist = T.get_targets(wdir).reset_index()
        options = [{"value": i, "label": i} for i in peaklist.peak_label]
        return options, options



import numpy as np

from dash import html, dcc
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from ms_mint.Mint import Mint
from ms_mint.plotly_tools import plotly_heatmap

from . import tools as T

_label = "Heatmap"


heatmap_options = [
    {"label": "Normalized by biomarker", "value": "normed_by_cols"},
    {"label": "Cluster", "value": "clustered"},
    {"label": "Dendrogram", "value": "add_dendrogram"},
    {"label": "Transposed", "value": "transposed"},
    {"label": "Correlation", "value": "correlation"},
    {"label": "Show in new tab", "value": "call_show"},
    {"label": "log1p", "value": "log1p"},
]


_layout = html.Div(
    [
        html.H3("Heatmap"),
        html.Button("Update", id="heatmap-update"),
        dcc.Dropdown(
            id="heatmap-options",
            value=["normed_by_cols"],
            options=heatmap_options,
            multi=True,
        ),
        dcc.Loading(
            html.Div(
                [dcc.Graph(id="heatmap-figure")],
                style={"height": "100vh", "marginTop": "50px"},
            ),
        ),
    ]
)


def layout():
    return _layout


def callbacks(app, fsc, cache):
    @app.callback(
        Output("heatmap-controls", "children"),
        Input("ana-secondary-tab", "value"),
        State("wdir", "children"),
    )
    def heat_controls(tab, wdir):
        if tab != _label:
            raise PreventUpdate
        return _layout

    @app.callback(
        Output("heatmap-figure", "figure"),
        Input("heatmap-update", "n_clicks"),
        State("ana-file-types", "value"),
        State("ana-peak-labels-include", "value"),
        State("ana-peak-labels-exclude", "value"),
        State("ana-ms-order", "value"),
        State("heatmap-options", "value"),
        State("viewport-container", "children"),
        State("wdir", "children"),
    )
    def heat_heatmap(
        n_clicks,
        file_types,
        include_labels,
        exclude_labels,
        ms_order,
        options,
        viewport,
        wdir,
    ):
        mint = Mint()

        width, height = [int(e) for e in viewport.split(",")]

        df = T.get_complete_results(
            wdir,
            include_labels=include_labels,
            exclude_labels=exclude_labels,
            file_types=file_types,
        )

        if len(df) == 0:
            return "No results yet. First run MINT."

        mint.results = df

        var_name = "peak_max"
        data = mint.crosstab(var_name)
        data.index = [T.Basename(i) for i in data.index]

        if ms_order is not None and len(ms_order) > 0:
            df = df.sort_values(ms_order)
            ms_files = df["MS-file"].drop_duplicates()
            data = data.loc[ms_files]

        data.fillna(0, inplace=True)

        name = var_name
        if "log1p" in options:
            data = data.apply(np.log1p)
            name = f"log( {var_name}+1 )"

        fig = plotly_heatmap(
            data,
            height=height,
            width=width,
            normed_by_cols="normed_by_cols" in options,
            transposed="transposed" in options,
            clustered="clustered" in options,
            add_dendrogram="add_dendrogram" in options,
            correlation="correlation" in options,
            call_show="call_show" in options,
            name=name,
        )

        return fig



from dash import html, dcc
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc

from matplotlib import pyplot as plt

import seaborn as sns

plt.rcParams["figure.autolayout"] = False
sns.set_context("paper")


from . import tools as T


_label = "Plotting"

_kinds = [
    "bar",
    "violin",
    "box",
    "count",
    "boxen",
    "scatter",
    "line",
    "strip",
    "swarm",
    "point",
    "hist",
    "kde",
    "ecdf"
]

_palettes = [
    "Accent",
    "Accent_r",
    "Blues",
    "Blues_r",
    "BrBG",
    "BrBG_r",
    "BuGn",
    "BuGn_r",
    "BuPu",
    "BuPu_r",
    "CMRmap",
    "CMRmap_r",
    "Dark2",
    "Dark2_r",
    "GnBu",
    "GnBu_r",
    "Greens",
    "Greens_r",
    "Greys",
    "Greys_r",
    "OrRd",
    "OrRd_r",
    "Oranges",
    "Oranges_r",
    "PRGn",
    "PRGn_r",
    "Paired",
    "Paired_r",
    "Pastel1",
    "Pastel1_r",
    "Pastel2",
    "Pastel2_r",
    "PiYG",
    "PiYG_r",
    "PuBu",
    "PuBuGn",
    "PuBuGn_r",
    "PuBu_r",
    "PuOr",
    "PuOr_r",
    "PuRd",
    "PuRd_r",
    "Purples",
    "Purples_r",
    "RdBu",
    "RdBu_r",
    "RdGy",
    "RdGy_r",
    "RdPu",
    "RdPu_r",
    "RdYlBu",
    "RdYlBu_r",
    "RdYlGn",
    "RdYlGn_r",
    "Reds",
    "Reds_r",
    "Set1",
    "Set1_r",
    "Set2",
    "Set2_r",
    "Set3",
    "Set3_r",
    "Spectral",
    "Spectral_r",
    "Wistia",
    "Wistia_r",
    "YlGn",
    "YlGnBu",
    "YlGnBu_r",
    "YlGn_r",
    "YlOrBr",
    "YlOrBr_r",
    "YlOrRd",
    "YlOrRd_r",
    "afmhot",
    "afmhot_r",
    "autumn",
    "autumn_r",
    "binary",
    "binary_r",
    "bone",
    "bone_r",
    "brg",
    "brg_r",
    "bwr",
    "bwr_r",
    "cividis",
    "cividis_r",
    "cool",
    "cool_r",
    "coolwarm",
    "coolwarm_r",
    "copper",
    "copper_r",
    "cubehelix",
    "cubehelix_r",
    "flag",
    "flag_r",
    "gist_earth",
    "gist_earth_r",
    "gist_gray",
    "gist_gray_r",
    "gist_heat",
    "gist_heat_r",
    "gist_ncar",
    "gist_ncar_r",
    "gist_rainbow",
    "gist_rainbow_r",
    "gist_stern",
    "gist_stern_r",
    "gist_yarg",
    "gist_yarg_r",
    "gnuplot",
    "gnuplot2",
    "gnuplot2_r",
    "gnuplot_r",
    "gray",
    "gray_r",
    "hot",
    "hot_r",
    "hsv",
    "hsv_r",
    "icefire",
    "icefire_r",
    "inferno",
    "inferno_r",
    "jet",
    "jet_r",
    "magma",
    "magma_r",
    "mako",
    "mako_r",
    "nipy_spectral",
    "nipy_spectral_r",
    "ocean",
    "ocean_r",
    "pink",
    "pink_r",
    "plasma",
    "plasma_r",
    "prism",
    "prism_r",
    "rainbow",
    "rainbow_r",
    "rocket",
    "rocket_r",
    "seismic",
    "seismic_r",
    "spring",
    "spring_r",
    "summer",
    "summer_r",
    "tab10",
    "tab10_r",
    "tab20",
    "tab20_r",
    "tab20b",
    "tab20b_r",
    "tab20c",
    "tab20c_r",
    "terrain",
    "terrain_r",
    "turbo",
    "turbo_r",
    "twilight",
    "twilight_r",
    "twilight_shifted",
    "twilight_shifted_r",
    "viridis",
    "viridis_r",
    "vlag",
    "vlag_r",
    "winter",
    "winter_r",
]

kind_options = [{"value": x, "label": x.capitalize()} for x in _kinds]

palette_options = [{"value": x, "label": x} for x in _palettes]

options = [
    {"label": "Rotate x-ticks", "value": "rot-x-ticks"},
    {"label": "Share x-axis", "value": "share-x"},
    {"label": "Share y-axes", "value": "share-y"},
    {"label": "Scientific notation", "value": "sci"},
    {"label": "Logarithmic x-scale", "value": "log-x"},
    {"label": "Logarithmic y-scale", "value": "log-y"},
    {"label": "High Quality", "value": "HQ"},
    {"label": "Don't dodge", "value": "no-dodge"},
]

_layout = html.Div(
    [
        html.H3(_label),
        dbc.Row([
            dbc.Col([
              
                html.Label("Figure kind"),
                dcc.Dropdown(id="plot-kind", options=kind_options, value="bar"),
                html.Label("Facet dimensions:"),
                dbc.Row([
                    dcc.Input( id="plot-fig-height", placeholder="Facet height", value=2.5, type="number"),
                    dcc.Input( id="plot-fig-aspect", placeholder="Facet aspect", value=5, type="number"),   
                ]),
                html.Label("Column wrap"),
                dcc.Slider(id="plot-col-wrap", step=1, min=0, max=30, value=0),                 
            ]),
            dbc.Col([                
                              
                html.Label("x- and y-axes"),
                dcc.Dropdown(id="plot-x", options=[{"value": "MS-file", "label": "MS-file"}], value="MS-file", placeholder="X"),
                dcc.Dropdown(id="plot-y", options=[{"value": "peak_area_top3", "label": "peak_area_top3"}], value="peak_area_top3", placeholder="Y"),  

                html.Label("Row and column facets:"),
                dcc.Dropdown(id="plot-col", options=[], value=None, placeholder="Columns"),
                dcc.Dropdown(id="plot-row", options=[{"value": "peak_label", "label": "peak_label"}], value="peak_label", placeholder="Rows"), 

            ]),
            dbc.Col([

                html.Label("Marker style and size:"),
                dcc.Dropdown(id="plot-style", options=[], value=None, placeholder="Style"),
                dcc.Dropdown(id="plot-size", options=[], value=None, placeholder="Size"),       
                         
                html.Label("Colors"),
                dcc.Dropdown(id="plot-hue", options=[], value=None, placeholder="Color"),
                dcc.Dropdown(id="plot-palette", options=palette_options, value=None, placeholder="Palette (Colors)",),
            ]),
            dbc.Col([                

                html.Label("Figure title"),
                dcc.Input(id="plot-title", placeholder="Figure title", value=None, style={'width': '100%'}),                  
            ]),
        ]),
        html.Label("Options"),
        dcc.Dropdown(id="plot-options", value=["sci"], options=options, multi=True),    
        html.Button("Update", id="plot-update"),
        dcc.Loading(
            html.Div(
                id="plot-figures",
                style={
                    "margin": "auto",
                    "marginTop": "10%",
                    "text-align": "center",
                    "maxWidth": "100%",
                    "minHeight": "300px",
                },
            )
        ),
    ]
)

_ouptuts = html.Div([])


def layout():
    return _layout


def callbacks(app, fsc, cache, cpu=None):
    @app.callback(
        Output("plot-x", "options"),
        Output("plot-y", "options"),
        Output("plot-col", "options"),
        Output("plot-row", "options"),
        Output("plot-hue", "options"),
        Output("plot-size", "options"),
        Output("plot-style", "options"),
        Input("ana-secondary-tab", "value"),
        State("wdir", "children"),
    )
    def fill_options(tab, wdir):
        if tab != _label:
            raise PreventUpdate
        results = T.get_complete_results(wdir)
        results = results.dropna(axis=1, how="all")
        cols = results.columns
        options = [{"value": x, "label": x} for x in cols]
        return [options] * 7

    @app.callback(
        Output("plot-figures", "children"),
        Input("plot-update", "n_clicks"),
        State("plot-kind", "value"),
        State("plot-fig-height", "value"),
        State("plot-fig-aspect", "value"),
        State("plot-x", "value"),
        State("plot-y", "value"),
        State("plot-hue", "value"),
        State("plot-col", "value"),
        State("plot-row", "value"),
        State("plot-col-wrap", "value"),
        State("plot-style", "value"),
        State("plot-size", "value"),
        State("plot-title", "value"),
        State("ana-file-types", "value"),
        State("ana-peak-labels-include", "value"),
        State("ana-peak-labels-exclude", "value"),
        State("ana-ms-order", "value"),
        State("plot-palette", "value"),
        State("plot-options", "value"),
        State("wdir", "children"),
    )
    def create_figure(
        n_clicks,
        kind,
        height,
        aspect,
        x,
        y,
        hue,
        col,
        row,
        col_wrap,
        style,
        size,
        title,
        file_types,
        include_labels,
        exclude_labels,
        ms_order,
        palette,
        options,
        wdir,
    ):

        if n_clicks is None:
            raise PreventUpdate
        if col_wrap == 0:
            col_wrap = None
        if col is None and row is None:
            col_wrap = None
        if height is None:
            height = 2.5
        if aspect is None:
            aspect = 1

        height = min(float(height), 100)
        height = max(height, 1)
        aspect = max(0.01, float(aspect))
        aspect = min(aspect, 100)

        df = T.get_complete_results(
            wdir,
            include_labels=include_labels,
            exclude_labels=exclude_labels,
            file_types=file_types,
        )

        df = df[(df.peak_n_datapoints > 0) & (df.peak_max > 0)]

        if ms_order is None or len(ms_order) == 0:
            df = df.sort_values(
                [i for i in [col, row, hue, x, y, style, size] if i is not None]
            )
        else:
            df = df.sort_values(ms_order)

        if hue is not None:
            df = df[df[hue].notna()]

        n_c, n_r = 1, 1
        if col is not None:
            n_c = len(df[col].drop_duplicates())
        if row is not None:
            n_r = len(df[row].drop_duplicates())

        if (n_c is not None) and (col_wrap is not None):
            n_c = n_c // col_wrap
            n_r = n_c % col_wrap
        del n_r

        if hue is not None and ((x is None) and (y is None)):
            x = hue

        if kind in ["scatter", "line"]:
            plot_func = sns.relplot
            kwargs = dict(
                facet_kws=dict(
                    sharex="share-x" in options,
                    sharey="share-y" in options,
                    legend_out=True,
                ),
                style=style,
                size=size,
            )
        elif kind in ["hist", "kde", "ecdf"]:
            plot_func = sns.displot
            kwargs = dict(
                facet_kws=dict(
                    sharex="share-x" in options,
                    sharey="share-y" in options,
                    legend_out=True,
                ),
            )            
        else:
            plot_func = sns.catplot
            kwargs = dict(
                sharex="share-x" in options,
                sharey="share-y" in options,
                dodge="no-dodge" not in options,
                facet_kws=dict(legend_out=True),
            )
        
        print(x, y, hue)

        try:
            g = plot_func(
                x=x,
                y=y,
                hue=hue,
                col=col,
                row=row,
                col_wrap=col_wrap,
                data=df,
                kind=kind,
                height=height,
                aspect=aspect,
                palette=palette,
                **kwargs
            )
        except Exception as e:
            return dbc.Alert(str(e), color="danger")

        g.fig.subplots_adjust(top=0.9)
        g.set_titles(col_template="{col_name}", row_template="{row_name}", y=1.05)

        if "log-x" in options:
            g.set(xscale="log")
        if "log-y" in options:
            g.set(yscale="log")

        notation = "sci" if "sci" in options else "plain"
        for ax in g.axes.flatten():
            try:
                ax.ticklabel_format(style=notation, scilimits=(0, 0), axis="x")
            except:
                pass
            try:
                ax.ticklabel_format(style=notation, scilimits=(0, 0), axis="y")
            except:
                pass

        if "rot-x-ticks" in options:
            g.set_xticklabels(rotation=90)

        if title is not None:
            g.fig.suptitle(title, y=1.01)

        g.tight_layout(w_pad=0)

        src = T.fig_to_src(dpi=300 if "HQ" in options else None)

        return html.Img(src=src, style={"maxWidth": "80%"})



import os
import shutil
import logging

from tqdm import tqdm

import numpy as np

import matplotlib

matplotlib.use("Agg")
from matplotlib import pyplot as plt
import seaborn as sns

import dash
from dash import html, dcc
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State, ALL

import dash_bootstrap_components as dbc

import plotly.graph_objects as go

from ms_mint import Mint

from . import tools as T


info_txt = """
Creating chromatograms from mzXML/mzML files can last 
a long time the first time. Try converting your files to 
_feather_ format first.'
"""

_label = "Optimization"

_layout = html.Div(
    [
        html.H3("Target Optimization"),
        dcc.Markdown("---"),
        html.H4("File selection"),
        dcc.Dropdown(
            id="pko-ms-selection",
            options=[
                {
                    "label": "Use selected files from metadata table (PeakOpt)",
                    "value": "peakopt",
                },
                {"label": "Use all files (may take a long time)", "value": "all"},
            ],
            value="peakopt",
            clearable=False,
            style={'max-width': '400px'}
        ),
        html.Div(
            id="experimental",
            children=[
                dcc.Markdown("---"),
                dbc.Row(
                    style={"visibility": "hidden", "height": "0px"},
                    children=[
                        html.H4("Process all peaks"),
                        html.Div(
                            [
                                html.Button(
                                    "Find largest peaks for all",
                                    id="pko-find-largest-peak-for-all",
                                    style={"width": "100%"},
                                ),
                                html.Label("Margin:"),
                                dcc.Slider(
                                    id="pko-margin",
                                    min=0.1,
                                    max=20,
                                    step=0.05,
                                    value=0.3,
                                ),
                                html.Label("", id="pko-margin-display"),
                            ],
                            style={
                                "width": "45%",
                                "display": "inline-block",
                                "margin": "auto",
                            },
                        ),
                        html.Div(
                            [
                                html.Button(
                                    "Remove low intensity peaks",
                                    id="pko-remove-low-intensity",
                                    style={"width": "100%"},
                                ),
                                html.Label("Threshold:"),
                                dcc.Input(id="pko-threshold", value="1e4"),
                            ],
                            style={
                                "width": "45%",
                                "display": "inline-block",
                                "margin": "auto",
                            },
                        ),
                        dcc.Markdown("---"),
                    ],
                ),
            ],
        ),
        html.H4("Peak previews"),
        html.Button("Update peak previews", id="pko-peak-preview"),
        html.Button("Regenerate all figures", id="pko-peak-preview-from-scratch"),
        html.Button("Detect rt_min,rt_max for all targets", id="pko-optimize-rt"),
        dcc.Markdown("---"),

        html.Div(
            id="pko-peak-preview-images",
            style={"maxHeight": "600px", "overflowX": "scroll", "padding": "auto"},

        ),
        dcc.Markdown("---"),
        html.Div(id="pko-controls"),
        dcc.Dropdown(id="pko-dropdown", options=[], value=None),
        dbc.Progress(
            id="pko-progress-bar",
            value=0,
            style={"marginBottom": "20px", "width": "100%"},
        ),

        html.Button("Set RT to current view", id="pko-set-rt"),
        html.Button("Detect rt_min,rt_max", id="pko-find-largest-peak"),
        html.Button("Confirm retention time", id="pko-confirm-rt"),
        html.Button("Remove Peak", id="pko-delete", style={"float": "right"}),

        dcc.Loading(dcc.Graph("pko-figure")),
        dcc.Checklist(
            id="pko-figure-options",
            options=[{"value": "log", "label": "Logarithmic y-scale"}],
            value=[],
        ),        
        html.Div(id="pko-image-clicked", style={'visibility': 'hidden'}),
        html.Div(
            children=[
                html.Button("<< Previous", id="pko-prev"),
                html.Button("Suggest", id="pko-suggest-next"),
                html.Button("Next >>", id="pko-next"),
            ],
            style={"text-align": "center", "margin": "auto", "marginTop": "2%"},
        ),
    ]
)


pko_layout_no_data = html.Div(
    [
        dcc.Markdown(
            """### No targets found.
    You did not generate a targets yet.
    """
        )
    ]
)


_outputs = html.Div(
    id="pko-outputs",
    children=[
        html.Div(id={"index": "pko-set-rt-output", "type": "output"}),
        html.Div(id={"index": "pko-confirm-rt-output", "type": "output"}),
        html.Div(
            id={"index": "pko-find-largest-peak-for-all-output", "type": "output"}
        ),
        html.Div(id={"index": "pko-find-largest-peak-output", "type": "output"}),
        html.Div(id={"index": "pko-delete-output", "type": "output"}),
        html.Div(id={"index": "pko-remove-low-intensity-output", "type": "output"}),
    ],
)


def layout():
    return _layout


def callbacks(app, fsc, cache, cpu=None):
    @app.callback(
        Output("pko-dropdown", "options"),
        Input("tab", "value"),
        Input({"index": "pko-delete-output", "type": "output"}, "children"),
        State("wdir", "children"),
        State("pko-dropdown", "options"),
    )
    def pko_controls(tab, peak_deleted, wdir, old_options):
        if tab != _label:
            raise PreventUpdate
        targets = T.get_targets(wdir)
        if targets is None:
            logging.warning("Peaklist is None")
            raise PreventUpdate
        options = [
            {"label": label, "value": i} for i, label in enumerate(targets.index)
        ]
        if options == old_options:
            raise PreventUpdate
        return options

    @app.callback(
        Output("pko-figure", "figure"),
        Input("pko-dropdown", "value"),
        Input("pko-figure-options", "value"),
        Input({"index": "pko-set-rt-output", "type": "output"}, "children"),
        Input("pko-dropdown", "options"),
        Input(
            {"index": "pko-find-largest-peak-for-all-output", "type": "output"},
            "children",
        ),
        Input({"index": "pko-find-largest-peak-output", "type": "output"}, "children"),
        Input({"index": "pko-confirm-rt-output", "type": "output"}, "children"),
        State("pko-ms-selection", "value"),
        State("pko-margin", "value"),
        State("wdir", "children"),
    )
    def pko_figure(
        peak_label_ndx,
        options,
        n_clicks,
        options_changed,
        find_largest_peak,
        find_largest_peak_single,
        rt_set,
        ms_selection,
        margin,
        wdir,
    ):
        fig = None
        if peak_label_ndx is None:
            raise PreventUpdate
        targets = T.get_targets(wdir).reset_index()
        if ms_selection == "peakopt":
            ms_files = T.get_ms_fns_for_peakopt(wdir)
        elif ms_selection == "all":
            ms_files = T.get_ms_fns(wdir)

        cols = ["mz_mean", "mz_width", "rt", "rt_min", "rt_max", "peak_label"]

        peak_label_ndx = peak_label_ndx % len(targets)
        mz_mean, mz_width, rt, rt_min, rt_max, label = targets.loc[peak_label_ndx, cols]

        if rt is np.isnan(rt):
            if (not np.isnan(rt_min)) and (not np.isnan(rt_max)):
                rt = np.mean([rt_min, rt_max])
        else:
            if (rt_min is None) or np.isnan(rt_min):
                rt_min = max(0, rt - margin)
            if (rt_max is None) or np.isnan(rt_max):
                rt_max = rt + margin

        if True or fig is None:
            fig = go.Figure()
            fig.layout.hovermode = "closest"
            fig.layout.xaxis.range = [rt_min, rt_max]

            fig.update_layout(
                yaxis_title="Intensity",
                xaxis_title="Scan Time [s]",
                xaxis=dict(rangeslider=dict(visible=True)),
            )
            fig.update_layout(title=label)
            if "log" in options:
                fig.update_yaxes(type="log")

        if not np.isnan(rt):
            fig.add_vline(rt)

        if (not np.isnan(rt_min)) and (not np.isnan(rt_max)):
            fig.add_vrect(
                x0=rt_min, x1=rt_max, line_width=0, fillcolor="green", opacity=0.1
            )

        n_files = len(ms_files)
        for i, fn in tqdm(enumerate(ms_files), total=n_files, desc="PKO-figure"):
            fsc.set("progress", int(100 * (i + 1) / n_files))

            name = os.path.basename(fn)
            name, _ = os.path.splitext(name)
            chrom = T.get_chromatogram(fn, mz_mean, mz_width, wdir)
            fig.add_trace(
                go.Scatter(x=chrom["scan_time"], y=chrom["intensity"], name=name)
            )
            fig.update_layout(showlegend=False)
            fig.update_layout(hoverlabel=dict(namelength=-1))
        return fig

    @app.callback(
        Output("pko-progress-bar", "value"),
        Input("pko-dropdown", "value"),
        State("pko-dropdown", "options"),
    )
    def set_progress(value, options):
        if (value is None) or (options is None):
            raise PreventUpdate
        progress = int(100 * (value + 1) / len(options))
        return progress

    @app.callback(
        Output(
            {"index": "pko-find-largest-peak-for-all-output", "type": "output"},
            "children",
        ),
        Input("pko-find-largest-peak-for-all", "n_clicks"),
        State("pko-ms-selection", "value"),
        State("pko-margin", "value"),
        State("wdir", "children"),
    )
    def pko_optimise_rt_min_max(n_clicks, ms_selection, margin, wdir):
        if n_clicks is None:
            raise PreventUpdate
    
        targets = T.get_targets(wdir)

        if ms_selection == "peakopt":
            ms_files = T.get_ms_fns_for_peakopt(wdir)
        elif ms_selection == "all":
            ms_files = T.get_ms_fns(wdir)

        mint = Mint()
        mint.targets = targets
        mint.ms_files = ms_files
        mint.opt.find_rt_min_max()       
        new_targets = mint.targets

        T.write_targets(new_targets, wdir)

        return dbc.Alert("Peak optimization done.", color="info")

    @app.callback(
        Output({"index": "pko-set-rt-output", "type": "output"}, "children"),
        Input("pko-set-rt", "n_clicks"),
        State("pko-dropdown", "value"),
        State("pko-figure", "figure"),
        State("wdir", "children"),
    )
    def pko_set_rt_min_max(n_clicks, peak_label, fig, wdir):
        if n_clicks is None:
            raise PreventUpdate
        rt_min, rt_max = fig["layout"]["xaxis"]["range"]
        rt_min, rt_max = np.round(rt_min, 4), np.round(rt_max, 4)
        T.update_targets(wdir, peak_label, rt_min, rt_max)
        return dbc.Alert(f"Set RT span to ({rt_min},{rt_max})", color="info")

    @app.callback(
        Output({"index": "pko-confirm-rt-output", "type": "output"}, "children"),
        Input("pko-confirm-rt", "n_clicks"),
        State("pko-dropdown", "value"),
        State("pko-figure", "figure"),
        State("wdir", "children"),
    )
    def pko_confirm_rt(n_clicks, peak_label, fig, wdir):
        if n_clicks is None:
            raise PreventUpdate

        rt_min, rt_max = fig["layout"]["xaxis"]["range"]
        rt_min, rt_max = np.round(rt_min, 4), np.round(rt_max, 4)

        image_label = f"{peak_label}_{rt_min}_{rt_max}"

        _, fn = T.get_figure_fn(
            kind="peak-preview", wdir=wdir, label=image_label, format="png"
        )

        rt = np.mean([rt_min, rt_max])

        T.update_targets(wdir, peak_label, rt=rt)

        if os.path.isfile(fn):
            os.remove(fn)

        return dbc.Alert(f"Set RT span to ({rt_min},{rt_max})", color="info")

    @app.callback(
        Output("pko-dropdown", "value"),
        Input("pko-prev", "n_clicks"),
        Input("pko-suggest-next", "n_clicks"),
        Input("pko-next", "n_clicks"),
        Input("pko-image-clicked", "children"),
        State("pko-dropdown", "value"),
        State("pko-dropdown", "options"),
        State("wdir", "children"),
    )
    def pko_prev_next_suggest(
        n_prev, n_suggest, n_next, image_clicked, value, options, wdir
    ):
        if (
            n_prev is None
            and n_next is None
            and image_clicked is None
            and n_suggest is None
        ):
            raise PreventUpdate

        prop_id = dash.callback_context.triggered[0]["prop_id"]

        if prop_id.startswith("pko-suggest"):

            targets = T.get_targets(wdir).reset_index()

            # If targets have missing values in RT columns
            # return index of first target with missing value
            if T.has_na(targets[["rt_min", "rt_max"]]):
                ndx_na_rt_min = targets[targets["rt_min"].isna()].index.to_list()
                ndx_na_rt_max = targets[targets["rt_max"].isna()].index.to_list()
                return min(ndx_na_rt_min + ndx_na_rt_max)

            rt_means = targets[["rt_min", "rt_max"]].mean(axis=1)

            peak_label_ndx = np.argmax((targets.rt - rt_means).abs())

            return peak_label_ndx

        if prop_id.startswith("pko-image-clicked"):
            for entry in options:
                if entry["label"] == image_clicked:
                    return entry["value"]
        elif value is None:
            return 0
        elif prop_id.startswith("pko-prev"):
            return (value - 1) % len(options)
        elif prop_id.startswith("pko-next"):
            return (value + 1) % len(options)

    @app.callback(
        Output("pko-peak-preview-images", "children"),
        Input("pko-peak-preview", "n_clicks"),
        Input("pko-peak-preview-from-scratch", "n_clicks"),
        State("pko-ms-selection", "value"),
        State("wdir", "children"),
    )
    def peak_preview(n_clicks, from_scratch, ms_selection, wdir):  # peak_opt, #set_rt,
        if n_clicks is None:
            raise PreventUpdate
        # reset updating after 5 attempts
        n_attempts = fsc.get(f"{wdir}-update-attempt")
        if n_attempts is None:
            n_attempts = 1
        elif n_attempts % 5:
            fsc.set(f"{wdir}-updating", False)

        # increment counter of attempts
        fsc.set(f"{wdir}-update-attempt", n_attempts + 1)

        if fsc.get(f"{wdir}-updating") is True:
            raise PreventUpdate

        fsc.set(f"{wdir}-updating", True)

        prop_id = dash.callback_context.triggered[0]["prop_id"]
        regenerate = prop_id.startswith("pko-peak-preview-from-scratch")
        if regenerate:
            image_path = os.path.join(wdir, "figures", "peak-preview")
            if os.path.isdir(image_path):
                shutil.rmtree(image_path)

        if ms_selection == "peakopt":
            ms_files = T.get_ms_fns_for_peakopt(wdir)
        elif ms_selection == "all":
            ms_files = T.get_ms_fns(wdir)
        else:
            assert False, ms_selection

        if len(ms_files) == 0:
            return dbc.Alert(
                'No files selected for peak optimization in Metadata tab. Please, select some files in column "PeakOpt".',
                color="warning",
            )
        else:
            logging.info(
                f"Using {len(ms_files)} files for peak preview. ({ms_selection})"
            )

        targets = T.get_targets(wdir)

        file_colors = T.file_colors(wdir)

        n_total = len(targets)

        sns.set_context("paper")
        images = []
        for i, (peak_label, row) in tqdm(enumerate(targets.iterrows()), total=n_total):
            fsc.set("progress", int(100 * (i + 1) / n_total))
            mz_mean, mz_width, rt, rt_min, rt_max = row[
                ["mz_mean", "mz_width", "rt", "rt_min", "rt_max"]
            ]

            if rt_min is None or np.isnan(rt_min):
                rt_min = 0
            if  rt_max is None or np.isnan(rt_max):
                rt_max = 1000

            image_label = f"{peak_label}_{rt_min}_{rt_max}"

            _, fn = T.get_figure_fn(
                kind="peak-preview", wdir=wdir, label=image_label, format="png"
            )

            if not os.path.isfile(fn) or regenerate:
                logging.info(f"Regenerating figure for {peak_label}")
                create_preview_peakshape(
                    ms_files,
                    mz_mean,
                    mz_width,
                    rt,
                    rt_min,
                    rt_max,
                    image_label,
                    wdir,
                    title=peak_label,
                    colors=file_colors,
                )

            if os.path.isfile(fn):
                src = T.png_fn_to_src(fn)
            else:
                src = None

            _id = {"index": peak_label, "type": "image"}
            image_id = f"image-{i}"
            images.append(
                html.A(
                    id=_id,
                    children=html.Img(
                        src=src, height=300, id=image_id, style={"margin": "10px"}
                    ),
                )
            )
            images.append(
                dbc.Tooltip(peak_label, target=image_id, style={"font-size": "50"})
            )
        fsc.set(f"{wdir}-updating", False)
        return images

    @app.callback(
        Output("pko-image-clicked", "children"),
        # Input needs brakets to make prevent_initital_call work
        [Input({"type": "image", "index": ALL}, "n_clicks")],
        prevent_initial_call=True,
    )
    def pko_image_clicked(ndx):
        if ndx is None or len(ndx) == 0:
            raise PreventUpdate
        ctx = dash.callback_context
        clicked = ctx.triggered[0]["prop_id"]
        clicked = clicked.replace('{"index":"', "")
        clicked = clicked.split('","type":')[0].replace("\\", "")
        if len(dash.callback_context.triggered) > 1:
            raise PreventUpdate
        return clicked

    @app.callback(
        Output({"index": "pko-delete-output", "type": "output"}, "children"),
        Input("pko-delete", "n_clicks"),
        State("pko-dropdown", "value"),
        State("wdir", "children"),
    )
    def plk_delete(n_clicks, peak_ndx, wdir):
        if n_clicks is None:
            raise PreventUpdate
        targets = T.get_targets(wdir).reset_index()
        peak_label = targets.loc[peak_ndx, "peak_label"]
        targets = targets.drop(peak_ndx, axis=0)
        T.write_targets(targets, wdir)
        return dbc.Alert(f"{peak_label} removed from targets.", color="info")

    @app.callback(
        Output({"index": "pko-find-largest-peak-output", "type": "output"}, "children"),
        Input("pko-find-largest-peak", "n_clicks"),
        State("pko-dropdown", "value"),
        State("pko-ms-selection", "value"),
        State("pko-margin", "value"),
        State("wdir", "children"),
    )
    def find_largest_peak(n_clicks, peak_label_ndx, ms_selection, margin, wdir):
        if n_clicks is None:
            raise PreventUpdate
        if peak_label_ndx is None:
            raise PreventUpdate
        targets = T.get_targets(wdir).reset_index()

        if ms_selection == "peakopt":
            ms_files = T.get_ms_fns_for_peakopt(wdir)
        elif ms_selection == "all":
            ms_files = T.get_ms_fns(wdir)

        peak_label = targets.at[peak_label_ndx, 'peak_label']
        
        mint = Mint()
        mint.targets = targets
        mint.ms_files = ms_files
        mint.opt.find_rt_min_max(peak_labels=[peak_label])       
        new_targets = mint.targets
        new_targets.to_csv(T.get_targets_fn(wdir), index=False)

        return dbc.Alert(
            f"Done optimize rt_min rt_max for {peak_label}", color="info"
        )

    @app.callback(
        Output(
            {"index": "pko-remove-low-intensity-output", "type": "output"}, "children"
        ),
        Input("pko-remove-low-intensity", "n_clicks"),
        State("pko-ms-selection", "value"),
        State("pko-threshold", "value"),
        State("wdir", "children"),
    )
    def remove_low_intensity_peaks(n_clicks, ms_selection, threshold, wdir):
        if n_clicks is None:
            raise PreventUpdate
        logging.info("Remove low intensity peaks.")
        targets = T.get_targets(wdir)

        if ms_selection == "peakopt":
            ms_files = T.get_ms_fns_for_peakopt(wdir)
        elif ms_selection == "all":
            ms_files = T.get_ms_fns(wdir)

        def set_progress(x):
            fsc.set("progress", x)

        mint = Mint(verbose=True, progress_callback=set_progress)

        tmp_targets = targets.reset_index().copy()

        tmp_targets["rt_min"] = tmp_targets.rt_min.fillna(0)
        tmp_targets["rt_max"] = tmp_targets.rt_max.fillna(100)

        mint.ms_files = ms_files
        mint.targets = tmp_targets
        mint.run()
        peak_labels = mint.results[
            mint.results.peak_max > float(threshold)
        ].peak_label.drop_duplicates()
        targets = targets[targets.index.isin(peak_labels)]
        T.write_targets(targets, wdir)
        return dbc.Alert("Low intensity peaks removed.", color="info")

    @app.callback(
        Output("pko-margin-display", "children"), Input("pko-margin", "value")
    )
    def display_slider_value(value):
        return "Value: {} [min]".format(value)


def create_preview_peakshape(
    ms_files, mz_mean, mz_width, rt, rt_min, rt_max, image_label, wdir, title, colors
):
    """Create peak shape previews."""
    plt.figure(figsize=(2.5, 1.5), dpi=30)
    y_max = 0
    for fn in ms_files:
        color = colors[T.filename_to_label(fn)]
        if color is None or color == "":
            color = "grey"
        fn_chro = T.get_chromatogram(fn, mz_mean, mz_width, wdir)
        fn_chro = fn_chro[
            (rt_min < fn_chro["scan_time"]) & (fn_chro["scan_time"] < rt_max)
        ]
        plt.plot(fn_chro["scan_time"], fn_chro["intensity"], lw=1, color=color)
        y_max = max(y_max, fn_chro["intensity"].max())
    if (not np.isnan(rt)) and not (np.isnan(rt_max)) and not (np.isnan(rt_min)):
        x = max(min(rt, rt_max), rt_min)
        rt_mean = np.mean([rt_min, rt_max])
        color_value = np.abs(rt_mean - rt)
        color = T.float_to_color(color_value, vmin=0, vmax=1, cmap="coolwarm")
        plt.vlines(x, 0, y_max, lw=3, color=color)
    plt.gca().set_title(title[:30], y=1.0, pad=15)
    plt.gca().ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
    plt.xlabel("Scan Time [s]")
    plt.ylabel("Intensity")
    filename = T.savefig(kind="peak-preview", wdir=wdir, label=image_label)
    plt.close()
    return filename



import logging
import dash

from dash import html
import dash_bootstrap_components as dbc

from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, ALL

_label = None

_layout = html.Div([html.Div(id="message")])


def layout():
    return _layout


def callbacks(app, fsc, cache):
    @app.callback(
        Output("message", "children"),
        [Input({"type": "output", "index": ALL}, "children")],
        prevent_initial_call=True,
    )
    def message(message):
        # print(f'Message: {message}')
        ctx = dash.callback_context
        message = ctx.triggered[0]
        if (message is None) or (message["value"] is None):
            # print(message)
            raise PreventUpdate
        _type = message["value"]["type"]
        _chil = message["value"]["props"]["children"]
        # print(f'Message ({_type}):', _chil)
        if "color" in message["value"]["props"].keys():
            color = message["value"]["props"]["color"]
        else:
            color = None
        if _type == "Alert":
            return dbc.Alert(_chil, color=color)
        else:
            logging.warning(f"Undefined message type: {_type}, {_chil}")
        return dbc.Alert("Undefined message", "danger")



import os
import logging

import pandas as pd

import dash
from dash import html, dcc
from dash.dependencies import Input, Output, State
from dash_tabulator import DashTabulator

import dash_bootstrap_components as dbc

from . import tools as T

from ms_mint.standards import TARGETS_COLUMNS


columns = [{"name": i, "id": i, "selectable": True} for i in TARGETS_COLUMNS]

tabulator_options = {
    "groupBy": "Label",
    "selectable": True,
    "headerFilterLiveFilterDelay": 3000,
    "layout": "fitDataFill",
    "height": "900px",
}

downloadButtonType = {
    "css": "btn btn-primary",
    "text": "Export",
    "type": "csv",
    "filename": "Targets",
}

clearFilterButtonType = {"css": "btn btn-outline-dark", "text": "Clear Filters"}

pkl_table = html.Div(
    id="pkl-table-container",
    style={"minHeight": 100, "margin": "50px 50px 0px 0px"},
    children=[
        DashTabulator(
            id="pkl-table",
            columns=T.gen_tabulator_columns(
                [
                    "peak_label",
                    "mz_mean",
                    "mz_width",
                    "rt",
                    "rt_min",
                    "rt_max",
                    "rt_unit",
                    "intensity_threshold",
                    "target_filename",
                ]
            ),
            options=tabulator_options,
            downloadButtonType=downloadButtonType,
            clearFilterButtonType=clearFilterButtonType,
        )
    ],
)

_label = "Targets"

_layout = html.Div(
    [
        html.H3(_label),
        dcc.Upload(
            id="pkl-upload",
            children=html.Div(["Drag and Drop or ", html.A("Select Files")]),
            style={
                "width": "100%",
                "height": "60px",
                "lineHeight": "60px",
                "borderWidth": "1px",
                "borderStyle": "dashed",
                "borderRadius": "5px",
                "textAlign": "center",
                "margin": "10px",
            },
            # Allow multiple files to be uploaded
            multiple=True,
        ),
        dcc.Dropdown(
            id="pkl-ms-mode",
            options=[
                {
                    "value": "positive",
                    "label": "Add proton mass to formula (positive mode)",
                },
                {
                    "value": "negative",
                    "label": "Subtract proton mass from formula (negative mode)",
                },
            ],
            value=None,
        ),
        html.Button("Save", id="pkl-save"),
        html.Button("Clear", id="pkl-clear", style={"float": "right"}),
        pkl_table,
    ]
)


_outputs = html.Div(
    id="pkl-outputs",
    children=[
        html.Div(id={"index": "pkl-upload-output", "type": "output"}),
        html.Div(id={"index": "pkl-save-output", "type": "output"}),
        html.Div(id={"index": "pkl-clear-output", "type": "output"}),
    ],
)


def layout():
    return _layout


def callbacks(app, fsc=None, cache=None):
    @app.callback(
        Output("pkl-table", "data"),
        Input("pkl-upload", "contents"),
        Input("pkl-ms-mode", "value"),
        Input("pkl-clear", "n_clicks"),
        State("pkl-upload", "filename"),
        State("pkl-upload", "last_modified"),
        State("wdir", "children"),
    )
    def pkl_upload(
        list_of_contents, ms_mode, clear, list_of_names, list_of_dates, wdir
    ):
        prop_id = dash.callback_context.triggered[0]["prop_id"]
        if prop_id.startswith("pkl-clear"):
            return pd.DataFrame(columns=TARGETS_COLUMNS).to_dict("records")
        target_dir = os.path.join(wdir, "target")
        fn = T.get_targets_fn(wdir)
        if list_of_contents is not None:
            # process data in drag n drop field
            dfs = [
                T.parse_pkl_files(c, n, d, target_dir, ms_mode=ms_mode)
                for c, n, d in zip(list_of_contents, list_of_names, list_of_dates)
            ]
            data = dfs[0].to_dict("records")
            return data
        elif os.path.isfile(fn):
            # read from harddrive
            return T.read_targets(fn).to_dict("records")
        else:
            logging.warning(f"Targets file not found: {fn}")

    @app.callback(
        Output({"index": "pkl-save-output", "type": "output"}, "children"),
        Input("pkl-save", "n_clicks"),
        Input("pkl-table", "data"),
        Input("pkl-table", "dataChanged"),
        State("wdir", "children"),
    )
    def plk_save(n_clicks, data, data_changed, wdir):
        df = pd.DataFrame(data)
        if len(df) == 0:
            df = pd.DataFrame(columns=TARGETS_COLUMNS)
        T.write_targets(df, wdir)
        return dbc.Alert("Peaklist saved.", color="info")

    @app.callback(
        Output("pkl-table", "downloadButtonType"),
        Input("tab", "value"),
        State("active-workspace", "children"),
    )
    def table_export_fn(tab, ws_name):
        fn = f"{T.today()}-{ws_name}_MINT-target"
        downloadButtonType = {
            "css": "btn btn-primary",
            "text": "Export",
            "type": "csv",
            "filename": fn,
        }
        return downloadButtonType
